unit Main;

{
  Inno Setup
  Copyright (C) 1998-2000 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  Background form
}

interface

{$I VERSION.INC}

uses
  WinTypes, WinProcs, SysUtils, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls,
  Struct;

const
  WM_SETUPNEXTSTEP = WM_USER + 3900;

type
  TMainForm = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormActivate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private declarations }
    procedure AppOnActivate (Sender: TObject);
    procedure AppOnRestore (Sender: TObject);
    function MainWindowHook (var Message: TMessage): Boolean;
    {$IFNDEF WIN32}
    procedure UpdateApplicationWindowPos;
    {$ENDIF}
    procedure WMSysCommand (var Message: TWMSysCommand); message WM_SYSCOMMAND;
    procedure WMSetupNextStep (var Msg: TMessage); message WM_SETUPNEXTSTEP;
    procedure WMEraseBkgnd (var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMGetDlgCode (var Message: TWMGetDlgCode); message WM_GETDLGCODE;
  protected
    function GetPalette: HPALETTE; override;
  public
    { Public declarations }
    CurStep: (csStart, csWizard, csCopy, csFinished, csTerminate);
    Closing, IsMinimized, InExecuteLoop: Boolean;
    NewPalette: HPALETTE;
    procedure CenterForm (AForm: TForm);
    procedure ShowAboutBox;
  end;

  TInstallOnThisVersionResult = (irInstall, irNotOnThisPlatform,
    irVerTooLow, irVerTooHigh);
  TShellFolderID = (sfDesktop, sfStartMenu, sfPrograms, sfStartup, sfSendTo,
    sfFonts, sfAppData, sfDocs);

var
  MainForm: TMainForm;

  SetupLdrMode: Boolean;
  SetupLdrOriginalFilename: String;
  SetupLdrOffsetMsg, SetupLdrOffset0, SetupLdrOffset1: Longint;

  InitDir, InitProgramGroup: String;

  SourceDir, TempInstallDir, WinDir, WinSystemDir, SystemDrive, FontDir: String;
  {$IFDEF WIN32}
  ProgramFilesDir, CommonFilesDir, DAODir: String;
  ShellFolders: array[Boolean, TShellFolderID] of String;
  {$ENDIF}

  ID: array[1..8] of Char;
  MinimumSpace: Longint;

  SetupHeader: TSetupHeader;

  { These Lists are allocated in this unit's initialization code. It doesn't
    bother freeing these Lists or their contents since this is done
    automatically when the application terminates }
  DirEntries,
  FileEntries,
  FileLocationEntries,
  IconEntries,
  IniEntries,
  RegistryEntries,
  InstallDeleteEntries,
  UninstallDeleteEntries,
  RunEntries,
  UninstallRunEntries: TList;

  DeleteFilesAfterInstallList, DeleteDirsAfterInstallList: TStringList;

  HasIcons, IsNT, NewGUI {$IFDEF WIN32}, IsAdmin {$ENDIF}: Boolean;
  WinVersion: Word;

  LicenseText, InfoBeforeText, InfoAfterText: PChar;

  WizardImage: TBitmap;

  NeedPassword, NeedsRestart, RestartSystem: Boolean;

function ChangeDirConst (const S: String): String;
function ChangeDirConstEx (const S: String; const CustomConsts: array of String): String;
function ChangeDirConstPChar (const S: PChar): String;
procedure DeinitSetup;
procedure InitializeSetup;
function InstallOnThisVersion (const MinVersion: TSetupVersionData;
  const OnlyBelowVersion: TSetupVersionData): TInstallOnThisVersionResult;
function ExitSetupMsgBox: Boolean;

implementation

uses
  Msgs, MsgIDs, ChildFrm, Wizard, Install, InstFunc,
  ShellAPI, {$IFDEF WIN32} ShlObj, {$ENDIF}
  CmnFunc, CmnFunc2,
  zlib, SetupEnt, Thunk;

{$R *.DFM}

procedure TMainForm.WMSysCommand (var Message: TWMSysCommand);
begin
  if (Message.CmdType = 9999) and not InExecuteLoop then
    ShowAboutBox
  else
    inherited;
end;

function TMainForm.GetPalette: HPALETTE;
begin
  Result := NewPalette;
end;

function ChangeDirConst (const S: String): String;
begin
  Result := ChangeDirConstEx(S, ['']);
end;

function ChangeDirConstEx (const S: String; const CustomConsts: array of String): String;

  {$IFDEF WIN32}
  function GetComputerNameString: String;
  var
    Buf: array[0..MAX_COMPUTERNAME_LENGTH] of Char;
    Size: DWORD;
  begin
    Size := SizeOf(Buf);
    if GetComputerName(Buf, Size) then
      Result := Buf
    else
      Result := '';
  end;
  {$ENDIF}

var
  Start, I, J, K: Integer;
  Cnst: String;
  Replaced: Boolean;

  procedure ReplaceConst (const ReplaceWith: String; const IsPath: Boolean);
  begin
    Replaced := True;
    Delete (Result, I, J-I+1);
    Insert (ReplaceWith, Result, I);
    Start := I + Length(ReplaceWith);
    if IsPath and (Length(ReplaceWith) <> 0) and
       (ReplaceWith[Length(ReplaceWith)] = '\') and
       (Start <= Length(Result)) and (Result[Start] = '\') then
      Delete (Result, Start, 1);
  end;

const
  FolderConsts: array[Boolean, TShellFolderID] of String =
    (('userdesktop', 'userstartmenu', 'userprograms', 'userstartup',
      'sendto', 'fonts', 'userappdata', 'userdocs'),
     ('commondesktop', 'commonstartmenu', 'commonprograms', 'commonstartup',
      'sendto', 'fonts', 'userappdata', 'userdocs'));
var
  B: Boolean;
  SF: TShellFolderID;
label 1;
begin
  Result := S;
  Start := 1;
1:for I := Start to Length(Result) do
    if Result[I] = '{' then
      for J := I+1 to Length(Result) do
        case Result[J] of
          '{': begin
                 Delete (Result, J, 1);
                 Start := J;
                 goto 1;
               end;
          '}': begin
                 Start := J+1;
                 Cnst := Copy(Result, I+1, J-I-1);
                 Replaced := False;
                 if Cnst <> '' then begin
                   if Cnst[1] = '%' then ReplaceConst (GetEnv(Copy(Cnst, 2, Maxint)), True) else
                   if Cnst = '\' then ReplaceConst ('\', False) else
                   if Cnst = 'src' then ReplaceConst (SourceDir, True) else
                   if Cnst = 'srcexe' then ReplaceConst (SetupLdrOriginalFilename, True) else
                   if Cnst = 'tmp' then ReplaceConst (TempInstallDir, True) else
                   if Cnst = 'app' then ReplaceConst (WizardDirValue, True) else
                   if Cnst = 'win' then ReplaceConst (WinDir, True) else
                   if Cnst = 'sys' then ReplaceConst (WinSystemDir, True) else
                   if Cnst = 'sd' then ReplaceConst (SystemDrive, True) else
                   if Cnst = 'hwnd' then ReplaceConst (IntToStr(MainForm.Handle), False) else
                   {$IFDEF WIN32}
                   if Cnst = 'pf' then ReplaceConst (ProgramFilesDir, True) else
                   if Cnst = 'cf' then ReplaceConst (CommonFilesDir, True) else
                   if Cnst = 'computername' then ReplaceConst (GetComputerNameString, False) else
                   if Cnst = 'dao' then ReplaceConst (DAODir, True) else
                   {$ENDIF}
                   if Cnst = 'groupname' then ReplaceConst (WizardGroupValue, False) else
                   if Cnst = 'group' then begin
                     {$IFDEF WIN32}
                     if ShellFolders[not(shAlwaysUsePersonalGroup in SetupHeader.Options), sfPrograms] = '' then
                     {$ENDIF}
                       raise Exception.Create(FmtSetupMessage1(msgErrorInternal, 'C100'));
                     {$IFDEF WIN32}
                     ReplaceConst (AddBackslash(ShellFolders[not(shAlwaysUsePersonalGroup in SetupHeader.Options),
                       sfPrograms]) + WizardGroupValue, True);
                     {$ENDIF}
                   end
                   else
                   for B := False to True do
                     for SF := Low(SF) to High(SF) do
                       if Cnst = FolderConsts[B, SF] then begin
                         {$IFDEF WIN32}
                         if ShellFolders[B, SF] = '' then begin
                         {$ENDIF}
                           if SF <> sfFonts then
                             raise Exception.Create(FmtSetupMessage1(msgErrorInternal, 'C101'))
                           else
                             ReplaceConst (FontDir, True);
                         {$IFDEF WIN32}
                         end
                         else
                           ReplaceConst (ShellFolders[B, SF], True);
                         {$ENDIF}
                         goto 1;
                       end;
                   K := 0;
                   while K < High(CustomConsts) do begin
                     if Cnst = CustomConsts[K] then
                       ReplaceConst (CustomConsts[K+1], False);
                     Inc (K, 2);
                   end;
                 end;
                 if not Replaced then
                   raise Exception.Create(FmtSetupMessage1(msgErrorInternal, 'C102'));
                 goto 1;
               end;
        end;
end;

function ChangeDirConstPChar (const S: PChar): String;
begin
  Result := ChangeDirConst(StrPas(S));
end;

function InstallOnThisVersion (const MinVersion: TSetupVersionData;
  const OnlyBelowVersion: TSetupVersionData): TInstallOnThisVersionResult;
var
  Ver, MinVer, OnlyBelowVer: Word;
begin
  Ver := WinVersion;
  if IsNT then begin
    MinVer := MinVersion.NTVersion;
    OnlyBelowVer := OnlyBelowVersion.NTVersion;
  end
  else begin
    MinVer := MinVersion.WinVersion;
    OnlyBelowVer := OnlyBelowVersion.WinVersion;
  end;
  Result := irInstall;
  if MinVer = 0 then
    Result := irNotOnThisPlatform
  else
    if Ver < MinVer then
      Result := irVerTooLow
    else
      if (OnlyBelowVer <> 0) and (Ver >= OnlyBelowVer) then
        Result := irVerTooHigh;
end;

function GetSizeOfFile (const Filename: String): Longint;
{ Returns size of Filename, or -1 if an error occured }
var
  {$IFDEF WIN32}
  H: THandle;
  FindData: TWin32FindData;
  {$ELSE}
  Rec: TSearchRec;
  {$ENDIF}
begin
  Result := -1;
  {$IFDEF WIN32}
  H := FindFirstFile(PChar(Filename), FindData);
  if H <> INVALID_HANDLE_VALUE then begin
    if FindData.nFileSizeHigh <> 0 then  { file size > 4GB ! }
      Result := High(Longint)  { just return maximum size of a Longint }
    else begin
      Result := FindData.nFileSizeLow;
      if Result < 0 then
        { if > 2GB (which is stored as a negative number on a signed 32-bit int), return 2GB }
        Result := High(Longint);
    end;
    WinProcs.FindClose (H);
  end;
  {$ELSE}
  if FindFirst(Filename, faHidden or faSysFile, Rec) = 0 then begin
    Result := Rec.Size;
    if Result < 0 then
      { if > 2GB (which is stored as a negative number on a signed 32-bit int), return 2GB }
      Result := High(Longint);
    FindClose (Rec);
  end;
  {$ENDIF}
end;

procedure InitializeSetup;
{ Initializes various vars used by the setup. This is called in the project
  source. }
  function VerToStr (Ver: Word): String;
  var
    Digits: Integer;
  begin
    with TSetupVersionDataVersion(Ver) do begin
      Digits := 2;
      if Minor mod 10 = 0 then begin
        Dec (Digits);
        Minor := Minor div 10;
      end;
      FmtStr (Result, '%d.%.*d', [Major, Digits, Minor]);
    end;
  end;
  procedure AbortInit (const Msg: TSetupMessageID);
  begin
    MsgBoxP (SetupMessages[Msg], nil, mbCriticalError, MB_OK);
    Abort;
  end;
  procedure AbortInitFmt1 (const Msg: TSetupMessageID; const Arg1: String);
  begin
    MsgBox (FmtSetupMessage(Msg, [Arg1]), '', mbCriticalError, MB_OK);
    Abort;
  end;
  procedure AbortInitVerError (const Msg: TSetupMessageID; const Platform: String;
    const Ver: Word);
  begin
    MsgBox (FmtSetupMessage(Msg, [Platform, VerToStr(Ver)]), '',
      mbCriticalError, MB_OK);
    Abort;
  end;
  {$IFDEF WIN32}
  function GetCommonShellFolderPath (const CommonFolderID, UserFolderID: Integer): String;
  begin
    if IsAdmin then begin
      Result := GetShellFolderPath(CommonFolderID);
      if Result <> '' then Exit;
    end;
    Result := GetShellFolderPath(UserFolderID);
  end;
  {$ENDIF}
  procedure ReadWizardImage (const Data: PDeflateBlockReadData);
  type
    PBuffer = ^TBuffer;
    TBuffer = array[0..8191] of Byte;
  var
    Buf: PBuffer;
    MemStream: TMemoryStream;
    BytesLeft, Bytes: Longint;
  begin
    Buf := nil;
    MemStream := nil;
    try
      New (Buf);
      MemStream := TMemoryStream.Create;
      InflateBlockRead (Data^, BytesLeft, SizeOf(BytesLeft));
      while BytesLeft > 0 do begin
        Bytes := BytesLeft;
        if Bytes > SizeOf(Buf^) then Bytes := SizeOf(Buf^);
        InflateBlockRead (Data^, Buf^, Bytes);
        MemStream.WriteBuffer (Buf^, Bytes);
        Dec (BytesLeft, Bytes);
      end;
      MemStream.Seek (0, soFromBeginning);
      WizardImage := TBitmap.Create;
      WizardImage.LoadFromStream (MemStream);
    finally
      MemStream.Free;
      if Assigned(Buf) then Dispose (Buf);
    end;
  end;
{$IFDEF WIN32}
const
  CSIDL_COMMON_STARTMENU = $0016;
  CSIDL_COMMON_PROGRAMS = $0017;
  CSIDL_COMMON_STARTUP = $0018;
  CSIDL_COMMON_DESKTOPDIRECTORY = $0019;
  CSIDL_APPDATA = $001A;
  FolderIDs: array[Boolean, TShellFolderID] of Integer =
    ((CSIDL_DESKTOPDIRECTORY, CSIDL_STARTMENU, CSIDL_PROGRAMS, CSIDL_STARTUP,
      CSIDL_SENDTO, CSIDL_FONTS, CSIDL_APPDATA, CSIDL_PERSONAL),
     (CSIDL_COMMON_DESKTOPDIRECTORY, CSIDL_COMMON_STARTMENU, CSIDL_COMMON_PROGRAMS, CSIDL_COMMON_STARTUP,
      CSIDL_SENDTO, CSIDL_FONTS, CSIDL_APPDATA, CSIDL_PERSONAL));
{$ENDIF}
var
  PCount, StartParam: Integer;

  SetupFilename: String;
  SetupFile: File;
  TestID: TSetupID;
  Data: PDeflateBlockReadData;
  BeganInflating: Boolean;
  NewEntry: Pointer;

  I: Integer;

  Name: String;
  F: File;

  J: TShellFolderID;

  CurFileSize: Longint;
  OldErrorMode: UINT;
begin
  { Based on SetupLdr or not?
    Parameters for launching SetupLdr-based installation are:
    /SL2 <original exe filename> <setup msg data offset>
         <setup 0 data offset> <setup 1 data offset>
  }
  PCount := NewParamCount;
  if (PCount >= 5) and (NewParamStr(1) = '/SL2') then begin
    StartParam := 6;
    SetupLdrMode := True;
    SetupLdrOriginalFilename := NewParamStr(2);
    SetupLdrOffsetMsg := StrToInt(NewParamStr(3));
    LoadCompressedSetupMessages (SetupLdrOriginalFilename, SetupLdrOffsetMsg,
      fmOpenRead or fmShareDenyWrite);
    SourceDir := ExtractFilePath(SetupLdrOriginalFilename);
    SetupLdrOffset0 := StrToInt(NewParamStr(4));
    SetupLdrOffset1 := StrToInt(NewParamStr(5));
  end
  else begin
    StartParam := 1;
    LoadCompressedSetupMessages (ExtractFilePath(ParamStr(0)) + 'SETUP.MSG', 0,
      fmOpenRead or fmShareDenyWrite);
    SetupLdrOriginalFilename := NewParamStr(0);
    SourceDir := ExtractFilePath(SetupLdrOriginalFilename);
  end;
  SourceDir := RemoveBackslashUnlessRoot(SourceDir);

  SetMessageBoxCaption (mbInformation, SetupMessages[msgInformationTitle]);
  SetMessageBoxCaption (mbConfirmation, SetupMessages[msgConfirmTitle]);
  SetMessageBoxCaption (mbError, SetupMessages[msgErrorTitle]);
  SetMessageBoxCaption (mbCriticalError, SetupMessages[msgErrorTitle]);

  {$IFDEF WIN32}
  if Win32Platform = VER_PLATFORM_WIN32s then
    AbortInitFmt1 (msgNotOnThisPlatform, 'Win32s');
  {$ENDIF}

  Application.Title := StrPas(SetupMessages[msgSetupAppTitle]);

  {$IFDEF WIN32}
  IsAdmin := IsAdminLoggedOn;
  {$ENDIF}
  Randomize;

  for I := StartParam to PCount do begin
    Name := NewParamStr(I);
    if CompareText(Copy(Name, 1, 5), '/DIR=') = 0 then
      InitDir := Copy(Name, 6, Maxint)
    else
    if CompareText(Copy(Name, 1, 7), '/GROUP=') = 0 then
      InitProgramGroup := Copy(Name, 8, Maxint);
  end;

  { Read SETUP.0, or from EXE }
  if not SetupLdrMode then begin
    SetupFilename := AddBackslash(SourceDir) + 'SETUP.0'; {don't localize}
    if not FileExists(SetupFilename) then
      AbortInitFmt1 (msgSetupFileMissing, 'SETUP.0');
  end
  else
    SetupFilename := SetupLdrOriginalFilename;
  AssignFile (SetupFile, SetupFilename);
  FileMode := fmOpenRead or fmShareDenyWrite;  Reset (SetupFile, 1);
  try
    Seek (SetupFile, SetupLdrOffset0);
    BlockRead (SetupFile, TestID, SizeOf(TestID));
    if TestID <> SetupID then
      AbortInit (msgSetupFileCorruptOrWrongVer);
    try
      BeganInflating := False;
      New (Data);
      try
        InflateBlockReadBegin (SetupFile, Data^);
        BeganInflating := True;
        { Header }
        SEInflateBlockRead (Data^, SetupHeader, SizeOf(SetupHeader),
          SetupHeaderStrings);
        case InstallOnThisVersion(SetupHeader.MinVersion, SetupHeader.OnlyBelowVersion) of
          irInstall: ;
          irNotOnThisPlatform:
            if IsNT then
              AbortInitFmt1 (msgNotOnThisPlatform, 'Windows NT')
            else
              AbortInitFmt1 (msgOnlyOnThisPlatform, 'Windows NT');
          irVerTooLow:
            if IsNT then
              AbortInitVerError (msgWinVersionTooLowError, 'Windows NT',
                SetupHeader.MinVersion.NTVersion)
            else
              AbortInitVerError (msgWinVersionTooLowError, 'Windows',
                SetupHeader.MinVersion.WinVersion);
          irVerTooHigh:
            if IsNT then
              AbortInitVerError (msgWinVersionTooHighError, 'Windows NT',
                SetupHeader.OnlyBelowVersion.NTVersion)
            else
              AbortInitVerError (msgWinVersionTooHighError, 'Windows',
                SetupHeader.OnlyBelowVersion.WinVersion);
        end;
        NeedPassword := shPassword in SetupHeader.Options;
        {$IFDEF WIN32}
        if (shAdminPrivilegesRequired in SetupHeader.Options) and
           not IsAdmin then
          AbortInit (msgAdminPrivilegesRequired);
        {$ENDIF}
        NeedsRestart := shAlwaysRestart in SetupHeader.Options;
        { Text }
        if SetupHeader.LicenseSize <> 0 then begin
          GetMem (LicenseText, SetupHeader.LicenseSize + 1{for null it adds});
          LicenseText[SetupHeader.LicenseSize] := #0; {add a null to the end}
          InflateBlockRead (Data^, LicenseText^, SetupHeader.LicenseSize);
        end;
        if SetupHeader.InfoBeforeSize <> 0 then begin
          GetMem (InfoBeforeText, SetupHeader.InfoBeforeSize + 1{for null it adds});
          InfoBeforeText[SetupHeader.InfoBeforeSize] := #0; {add a null to the end}
          InflateBlockRead (Data^, InfoBeforeText^, SetupHeader.InfoBeforeSize);
        end;
        if SetupHeader.InfoAfterSize <> 0 then begin
          GetMem (InfoAfterText, SetupHeader.InfoAfterSize + 1{for null it adds});
          InfoAfterText[SetupHeader.InfoAfterSize] := #0; {add a null to the end}
          InflateBlockRead (Data^, InfoAfterText^, SetupHeader.InfoAfterSize);
        end;
        { Wizard image }
        ReadWizardImage (Data);
        { Dir entries }
        for I := 1 to SetupHeader.NumDirEntries do begin
          GetMem (NewEntry, SizeOf(TSetupDirEntry));
          SEInflateBlockRead (Data^, NewEntry^, SizeOf(TSetupDirEntry),
            SetupDirEntryStrings);
          if InstallOnThisVersion(PSetupDirEntry(NewEntry)^.MinVersion,
             PSetupDirEntry(NewEntry)^.OnlyBelowVersion) = irInstall then
            DirEntries.Add (NewEntry)
          else
            SEFreeRec (NewEntry, SizeOf(TSetupDirEntry), SetupDirEntryStrings);
        end;
        { File entries }
        for I := 1 to SetupHeader.NumFileEntries do begin
          GetMem (NewEntry, SizeOf(TSetupFileEntry));
          SEInflateBlockRead (Data^, NewEntry^, SizeOf(TSetupFileEntry),
            SetupFileEntryStrings);
          if InstallOnThisVersion(PSetupFileEntry(NewEntry)^.MinVersion,
             PSetupFileEntry(NewEntry)^.OnlyBelowVersion) = irInstall then
            FileEntries.Add (NewEntry)
          else
            SEFreeRec (NewEntry, SizeOf(TSetupFileEntry), SetupFileEntryStrings);
        end;
        { Icon entries }
        for I := 1 to SetupHeader.NumIconEntries do begin
          GetMem (NewEntry, SizeOf(TSetupIconEntry));
          SEInflateBlockRead (Data^, NewEntry^, SizeOf(TSetupIconEntry),
            SetupIconEntryStrings);
          if InstallOnThisVersion(PSetupIconEntry(NewEntry)^.MinVersion,
             PSetupIconEntry(NewEntry)^.OnlyBelowVersion) = irInstall then
            IconEntries.Add (NewEntry)
          else
            SEFreeRec (NewEntry, SizeOf(TSetupIconEntry), SetupIconEntryStrings);
        end;
        { INI entries }
        for I := 1 to SetupHeader.NumIniEntries do begin
          GetMem (NewEntry, SizeOf(TSetupIniEntry));
          SEInflateBlockRead (Data^, NewEntry^, SizeOf(TSetupIniEntry),
            SetupIniEntryStrings);
          if InstallOnThisVersion(PSetupIniEntry(NewEntry)^.MinVersion,
             PSetupIniEntry(NewEntry)^.OnlyBelowVersion) = irInstall then
            IniEntries.Add (NewEntry)
          else
            SEFreeRec (NewEntry, SizeOf(TSetupIniEntry), SetupIniEntryStrings);
        end;
        { Registry entries }
        for I := 1 to SetupHeader.NumRegistryEntries do begin
          GetMem (NewEntry, SizeOf(TSetupRegistryEntry));
          SEInflateBlockRead (Data^, NewEntry^, SizeOf(TSetupRegistryEntry),
            SetupRegistryEntryStrings);
          if InstallOnThisVersion(PSetupRegistryEntry(NewEntry)^.MinVersion,
             PSetupRegistryEntry(NewEntry)^.OnlyBelowVersion) = irInstall then
            RegistryEntries.Add (NewEntry)
          else
            SEFreeRec (NewEntry, SizeOf(TSetupRegistryEntry), SetupRegistryEntryStrings);
        end;
        { InstallDelete entries }
        for I := 1 to SetupHeader.NumInstallDeleteEntries do begin
          GetMem (NewEntry, SizeOf(TSetupDeleteEntry));
          SEInflateBlockRead (Data^, NewEntry^, SizeOf(TSetupDeleteEntry),
            SetupDeleteEntryStrings);
          if InstallOnThisVersion(PSetupDeleteEntry(NewEntry)^.MinVersion,
             PSetupDeleteEntry(NewEntry)^.OnlyBelowVersion) = irInstall then
            InstallDeleteEntries.Add (NewEntry)
          else
            SEFreeRec (NewEntry, SizeOf(TSetupDeleteEntry), SetupDeleteEntryStrings);
        end;
        { UninstallDelete entries }
        for I := 1 to SetupHeader.NumUninstallDeleteEntries do begin
          GetMem (NewEntry, SizeOf(TSetupDeleteEntry));
          SEInflateBlockRead (Data^, NewEntry^, SizeOf(TSetupDeleteEntry),
            SetupDeleteEntryStrings);
          if InstallOnThisVersion(PSetupDeleteEntry(NewEntry)^.MinVersion,
             PSetupDeleteEntry(NewEntry)^.OnlyBelowVersion) = irInstall then
            UninstallDeleteEntries.Add (NewEntry)
          else
            SEFreeRec (NewEntry, SizeOf(TSetupDeleteEntry), SetupDeleteEntryStrings);
        end;
        { Run entries }
        for I := 1 to SetupHeader.NumRunEntries do begin
          GetMem (NewEntry, SizeOf(TSetupRunEntry));
          SEInflateBlockRead (Data^, NewEntry^, SizeOf(TSetupRunEntry),
            SetupRunEntryStrings);
          if InstallOnThisVersion(PSetupRunEntry(NewEntry)^.MinVersion,
             PSetupRunEntry(NewEntry)^.OnlyBelowVersion) = irInstall then
            RunEntries.Add (NewEntry)
          else
            SEFreeRec (NewEntry, SizeOf(TSetupRunEntry), SetupRunEntryStrings);
        end;
        { UninstallRun entries }
        for I := 1 to SetupHeader.NumUninstallRunEntries do begin
          GetMem (NewEntry, SizeOf(TSetupRunEntry));
          SEInflateBlockRead (Data^, NewEntry^, SizeOf(TSetupRunEntry),
            SetupRunEntryStrings);
          if InstallOnThisVersion(PSetupRunEntry(NewEntry)^.MinVersion,
             PSetupRunEntry(NewEntry)^.OnlyBelowVersion) = irInstall then
            UninstallRunEntries.Add (NewEntry)
          else
            SEFreeRec (NewEntry, SizeOf(TSetupRunEntry), SetupRunEntryStrings);
        end;

        BeganInflating := False;
        InflateBlockReadEnd (Data^);

        InflateBlockReadBegin (SetupFile, Data^);
        BeganInflating := True;

        { File location entries }
        for I := 1 to SetupHeader.NumFileLocationEntries do begin
          GetMem (NewEntry, SizeOf(TSetupFileLocationEntry));
          InflateBlockRead (Data^, NewEntry^, SizeOf(TSetupFileLocationEntry));
          FileLocationEntries.Add (NewEntry);
        end;
      finally
        if BeganInflating then
          InflateBlockReadEnd (Data^);
        Dispose (Data);
      end;
    except
      on EZlibDataError do
        AbortInit (msgSetupFileCorrupt);
    end;
  finally
    CloseFile (SetupFile);
  end;

  HasIcons := (IconEntries.Count <> 0) or ((shUninstallable in SetupHeader.Options)
    {$IFDEF WIN32} and ((shAlwaysCreateUninstallIcon in SetupHeader.Options) or not UsingWindows4) {$ENDIF});

  { Read Windows dir }
  WinDir := GetWinDir;

  { Read Windows System dir. If it is a shared version it uses Windows dir
    instead. }
  WinSystemDir := GetSystemDir;
  if Pos(WinDir, WinSystemDir) <> 1 then begin
    { If System is not a subdirectory of Windows, possibly a shared version }
    try
      { Try making a temp file to see if the dir is write protected }
      Name := GenerateUniqueName(WinSystemDir, '.tmp');
      AssignFile (F, Name);
      FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (F, 1);
      CloseFile (F);
      DeleteFile (Name);
    except
      { If exception occured, system dir is probably write protected. }
      WinSystemDir := WinDir;
    end;
  end;

  { Get system drive }
  {$IFDEF WIN32}
  if Win32Platform = VER_PLATFORM_WIN32_NT then
    SystemDrive := GetEnv('SystemDrive')  {don't localize}
  else
    SystemDrive := '';
  if SystemDrive = '' then begin
  {$ENDIF}
    SystemDrive := ExtractFileDrive(WinDir);
    if SystemDrive = '' then
      { In some rare case that ExtractFileDrive failed, just default to C }
      SystemDrive := 'C:';
  {$IFDEF WIN32}
  end;
  {$ENDIF}

  { Get Program Files, Common Files, and shell folder dirs on Win32 }
  {$IFDEF WIN32}
  ProgramFilesDir := GetProgramFilesPath;
  if ProgramFilesDir = '' then
    ProgramFilesDir := SystemDrive + '\Program Files';  {don't localize}
  CommonFilesDir := GetCommonFilesPath;
  if CommonFilesDir = '' then
    CommonFilesDir := AddBackslash(ProgramFilesDir) + 'Common Files';  {don't localize}
  if UsingWindows4 then begin
    for J := Low(J) to High(J) do
      ShellFolders[False, J] := RemoveBackslashUnlessRoot(GetShellFolderPath(FolderIDs[False, J]));
    for J := Low(J) to High(J) do
      ShellFolders[True, J] := RemoveBackslashUnlessRoot(GetCommonShellFolderPath(FolderIDs[True, J],
        FolderIDs[False, J]));
  end;
  {$ENDIF}

  { Get Fonts directory }
  {$IFDEF WIN32}
  FontDir := ShellFolders[False, sfFonts];
  if FontDir = '' then
  {$ENDIF}
    FontDir := WinSystemDir;

  { Generate DAO directory name }
  {$IFDEF WIN32}
  if UsingWindows4 then
    DAODir := AddBackslash(CommonFilesDir) + 'Microsoft Shared\DAO'
  else
    DAODir := AddBackslash(WinDir) + 'MSAPPS\DAO';
  {$ENDIF}

  { Attempt to create a TempInstallDir }
  Name := GenerateUniqueName(GetTempDir, '.tmp');
  {$IFNDEF WIN32}
  try
    MkDir (Name);
  except
    raise Exception.Create(FmtSetupMessage1(msgErrorCreatingDir, Name));
  end;
  {$ELSE}
  { More optimized way }
  if not CreateDirectory(PChar(Name), nil) then
    raise Exception.Create(FmtSetupMessage1(msgErrorCreatingDir, Name));
  {$ENDIF}
  TempInstallDir := Name;

  { After the rest of the variables are initialized, calculate minimum disk
    space it needs by adding all of the file's sizes.
    Also for each "external" file, check the file size now, and store it the
    ExternalSize field of the TSetupFileEntry record. }
  OldErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);  { Prevent "Network Error" boxes }
  try
    MinimumSpace := SetupHeader.ExtraDiskSpaceRequired;
    for I := 0 to FileEntries.Count-1 do
      with PSetupFileEntry(FileEntries[I])^ do
        if LocationEntry <> -1 then  { not an "external" file }
          Inc (MinimumSpace, PSetupFileLocationEntry(FileLocationEntries[
            LocationEntry])^.OriginalSize)
        else begin
          try
            CurFileSize := GetSizeOfFile(ChangeDirConstPChar(SourceFilename));
          except
            { ignore exceptions }
            CurFileSize := 0;
          end;
          if CurFileSize < 0 then
            CurFileSize := 0;
          ExternalSize := CurFileSize;
          Inc (MinimumSpace, CurFileSize);
        end;
  finally
    SetErrorMode (OldErrorMode);
  end;
end;

procedure TMainForm.AppOnActivate(Sender: TObject);
begin
  { This corrects some rare palette problems }
  if NewPalette <> 0 then
    Repaint;
end;

function ScaleRGB (const Color: TColor; const Scale: Byte): TColor;
type
  TColorBytes = array[0..3] of Byte;
var
  I: Integer;
begin
  Result := Color;
  for I := 0 to 2 do
    TColorBytes(Result)[I] := (TColorBytes(Result)[I] * Scale) div 255;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  SystemMenu: HMenu;
  Palette: packed record
    palVersion: Word;
    palNumEntries: Word;
    palPalEntry: array[0..63] of TPaletteEntry;
  end;
  Entry: Integer;
  C: TColor;
begin
  if shWindowVisible in SetupHeader.Options then begin
    { Should the main window not be sizable? }
    if not(shWindowShowCaption in SetupHeader.Options) then
      BorderStyle := bsNone
    else
    if not(shWindowResizable in SetupHeader.Options) then
      BorderStyle := bsSingle;

    { Make the main window full-screen }
    SetBounds (0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN));
    { Before maximizing the window, ensure Handle is created now so the correct
      'restored' position is saved properly }
    HandleNeeded;

    { Maximize the window so that the taskbar is still accessible }
    if shWindowStartMaximized in SetupHeader.Options then
      WindowState := wsMaximized;
  end
  else begin
    BorderStyle := bsNone;
    SetBounds (0, 0, 0, 0);
  end;

  Caption := FmtSetupMessage1(msgSetupWindowTitle, StrPas(SetupHeader.AppName));

  { Append the 'About Setup' item to the system menu }
  SystemMenu := GetSystemMenu(Handle, False);
  AppendMenu (SystemMenu, MF_SEPARATOR, 0, nil);
  AppendMenu (SystemMenu, MF_STRING, 9999, SetupMessages[msgAboutSetupMenuItem]);

  if not(shBackSolid in SetupHeader.Options) then begin
    { Create the palette }
    with Palette do begin
      palVersion := $300;
      palNumEntries := 64;
    end;
    C := ColorToRGB(SetupHeader.BackColor);
    for Entry := 0 to 63 do
      { rest of entries }
      TColor(Palette.palPalEntry[Entry]) := ScaleRGB(C, (Entry*4) xor 255);
    NewPalette := CreatePalette(PLogPalette(@Palette)^);
  end;

  Application.OnActivate := AppOnActivate;
  Application.HookMainWindow (MainWindowHook);
  Application.OnRestore := AppOnRestore;

  {$IFNDEF WIN32}
  UpdateApplicationWindowPos;
  {$ENDIF}
end;

procedure TMainForm.CenterForm (AForm: TForm);
var
  R: TRect;
  X, Y: Integer;
begin
  if shWindowVisible in SetupHeader.Options then begin
    R := ClientRect;
    MapWindowPoints (Handle, 0, R, 2);
  end
  else
    R := Rect(0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN));
  X := ((R.Right+R.Left) div 2)-(AForm.Width div 2);
  if X < 0 then X := 0;
  Y := ((R.Bottom+R.Top) div 2)-(AForm.Height div 2);
  if Y < 0 then Y := 0;
  AForm.SetBounds (X, Y, AForm.Width, AForm.Height);
end;

{$IFNDEF WIN32}
procedure TMainForm.UpdateApplicationWindowPos;
begin
  { Fixes the weird-looking minimizing seen in previous versions when
    window animation is enabled. This is used in combination with the fix
    in setup.dpr }
  SetWindowPos (Application.Handle, 0, Left, Top, Width, 0,
    SWP_NOACTIVATE or SWP_NOZORDER);
end;
{$ENDIF}

procedure TMainForm.WMEraseBkgnd (var Message: TWMEraseBkgnd);
begin
  { Since the form paints its entire client area in FormPaint, there is
    no need for the VCL to ever erase the client area with the brush color.
    Doing so only slows it down, so this message handler disables that default
    behavior. }
  Message.Result := 0;
end;

procedure TMainForm.FormPaint(Sender: TObject);
var
  C: TColor;
  CS: TPoint;
  Z: Integer;
  R, R2: TRect;
begin
  with Canvas do begin
    { Draw the blue background }
    if NewPalette <> 0 then
      SelectPalette (Handle, NewPalette, False);
    if shBackSolid in SetupHeader.Options then begin
      Brush.Color := SetupHeader.BackColor;
      FillRect (ClientRect);
    end
    else begin
      C := ColorToRGB(SetupHeader.BackColor);
      CS := ClientRect.BottomRight;
      for Z := 0 to 63 do begin
        Brush.Color := ScaleRGB(C, (Z*4) xor 255) or $02000000;
        FillRect (Rect(0, MulDiv(CS.Y, Z, 63), CS.X, MulDiv(CS.Y, Z+1, 63)));
      end;
    end;

    { Draw the application name and copyright }
    SetBkMode (Handle, TRANSPARENT);

    Font.Name := 'Arial';
    Font.Size := 29;
    Font.Style := [fsBold, fsItalic];
    R := ClientRect;
    InflateRect (R, -8, -8);
    R2 := R;
    OffsetRect (R2, 4, 4);
    Font.Color := clBlack;
    DrawText (Handle, SetupHeader.AppName, -1, R2, DT_TOP or DT_LEFT or
      DT_WORDBREAK or DT_NOPREFIX);
    Font.Color := clWhite;
    DrawText (Handle, SetupHeader.AppName, -1, R, DT_TOP or DT_LEFT or
      DT_WORDBREAK or DT_NOPREFIX);

    Font.Size := 8;
    Font.Style := [];
    R := ClientRect;
    InflateRect (R, -6, -6);
    R2 := R;
    DrawText (Handle, SetupHeader.AppCopyright, -1, R2, DT_TOP or DT_RIGHT or
      DT_WORDBREAK or DT_NOPREFIX or DT_CALCRECT);
    R.Top := R.Bottom - (R2.Bottom - R2.Top);
    R2 := R;
    OffsetRect (R2, 1, 1);
    Font.Color := clBlack;
    DrawText (Handle, SetupHeader.AppCopyright, -1, R2, DT_TOP or DT_RIGHT or
      DT_WORDBREAK or DT_NOPREFIX);
    Font.Color := clWhite;
    DrawText (Handle, SetupHeader.AppCopyright, -1, R, DT_TOP or DT_RIGHT or
      DT_WORDBREAK or DT_NOPREFIX);
  end;
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  {$IFNDEF WIN32}
  UpdateApplicationWindowPos;
  {$ENDIF}

  { Needs to redraw the background whenever the form is resized }
  Repaint;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Application.UnhookMainWindow (MainWindowHook);
  if NewPalette <> 0 then
    DeleteObject (NewPalette);
end;

procedure TMainForm.ShowAboutBox;
const
  DefAboutText = '%1 version %2' + SNewLine + '%3' + SNewLine2 +
    '%1 home page:' + SNewLine + '%4';
  AboutTextArg1 = 'Inno Setup';
  AboutTextArg2 = SetupVersion;
  AboutTextArg3 = 'Copyright (C) 1998-2000 Jordan Russell';
  AboutTextArg4 = 'http://www.jordanr.cjb.net/';
var
  P, Z: PChar;
  I: Integer;
  {$IFNDEF WIN32}
  Buf: array[0..4095] of Char;
  {$ELSE}
  S: String;
  {$ENDIF}
begin
  P := SetupMessages[msgAboutSetupMessage];
  I := 0;
  Z := P;
  while Z^ <> #0 do begin
    if Z^ = '%' then begin
      Inc (Z);
      if Z^ = #0 then Break;
      if Z^ in ['1'..'9'] then
        I := I or (1 shl (Ord(Z^) - Ord('1')));
    end;
    Inc (Z);
  end;
  if I and 15 <> 15 then
    P := DefAboutText;
  {$IFNDEF WIN32}
  StrFmtMessage (Buf, SizeOf(Buf)-1, P, [AboutTextArg1, AboutTextArg2,
    AboutTextArg3, AboutTextArg4]);
  if SetupMessages[msgAboutSetupNote, 0] <> #0 then begin
    StrLCat (Buf, SNewLine2, SizeOf(Buf)-1);
    StrLCat (Buf, SetupMessages[msgAboutSetupNote], SizeOf(Buf)-1);
  end;
  MsgBoxP (Buf, SetupMessages[msgAboutSetupTitle], mbInformation, MB_OK);
  {$ELSE}
  S := FmtMessage(P, [AboutTextArg1, AboutTextArg2, AboutTextArg3,
    AboutTextArg4]);
  if SetupMessages[msgAboutSetupNote, 0] <> #0 then
    S := S + SNewLine2 + SetupMessages[msgAboutSetupNote];
  MsgBoxP (PChar(S), SetupMessages[msgAboutSetupTitle], mbInformation, MB_OK);
  {$ENDIF}
end;

var
  ReadmeExecuteName: String;

procedure ProcessMessagesProc; far;
begin
  Application.ProcessMessages;
end;

procedure TMainForm.WMSetupNextStep (var Msg: TMessage);
var
  I, ErrorCode: Integer;
  S: String;
label NextStep;
begin
  NextStep:
  { Finish up the previous step }
  case CurStep of
    csStart: ;
    csWizard: begin
        WizardDirValue := WizardForm.DirEdit.Text;
        WizardGroupValue := WizardForm.GroupEdit.Text;
        WizardNoIcons := WizardForm.NoIconsCheck.Checked;
      end;
    csCopy: begin
        InstallForm.Free;
        InstallForm := nil;
        if RunEntries.Count <> 0 then begin
          InExecuteLoop := True;
          try
            for I := 0 to RunEntries.Count-1 do
              with PSetupRunEntry(RunEntries[I])^ do begin
                S := ChangeDirConstPChar(Name);
                if not(roShellExec in Options) then begin
                  if not InstExec(S, ChangeDirConstPChar(Parameters),
                     ChangeDirConstPChar(WorkingDir), Wait = rwWaitUntilTerminated,
                     Wait = rwWaitUntilIdle, ProcessMessagesProc, ErrorCode) then
                    MsgBox (FmtSetupMessage1(msgErrorExecutingProgram, S) + SNewLine2 +
                      {$IFDEF WIN32}
                      FmtSetupMessage(msgErrorFunctionFailedWithMessage, ['CreateProcess',
                        IntToStr(ErrorCode), SysErrorMessage(ErrorCode)]),
                      {$ELSE}
                      FmtSetupMessage(msgErrorFunctionFailed, ['WinExec',
                        IntToStr(ErrorCode)]),
                      {$ENDIF}
                      '', mbCriticalError, MB_OK);
                end
                else begin
                  if not InstShellExec(S, ChangeDirConstPChar(Parameters),
                     ChangeDirConstPChar(WorkingDir), ErrorCode) then
                    MsgBox (FmtSetupMessage1(msgErrorExecutingProgram, S) + SNewLine2 +
                      FmtSetupMessage(msgErrorFunctionFailed, ['ShellExecute',
                        IntToStr(ErrorCode)]),
                      '', mbCriticalError, MB_OK);
                end;
              end;
          finally
            InExecuteLoop := False;
          end;
          Application.BringToFront;
        end;
      end;
    csFinished: begin
        if not NeedsRestart then begin
          if (ReadmeExecuteName <> '') and WizardForm.ShowReadmeCheck.Checked then
            { note: changed from nil to 'open' for compatibility with a
              non-English version of Windows }
            if ShellExecute(Handle, 'open', StringAsPChar(ReadmeExecuteName), nil,
               nil, SW_SHOW) <= 32 then
              MsgBoxP (SetupMessages[msgErrorOpeningReadme], nil, mbCriticalError, MB_OK);
        end
        else
          RestartSystem := WizardForm.YesRadio.Checked;
      end;
  end;

  { Proceed to next step }
  Inc (CurStep);
  case CurStep of
    csWizard: begin
          Application.Restore;
          WizardForm := TWizardForm.Create(Application);
          WizardForm.Show;
        end;
    csCopy: begin
          Application.Restore;
          Update;
          InstallForm := TInstallForm.Create(Application);
          InstallForm.Show;
          InstallForm.StartInstalling;
        end;
    csFinished: begin
          if not NeedsRestart then begin
            for I := 0 to FileEntries.Count-1 do
              with PSetupFileEntry(FileEntries[I])^ do
                if foIsReadmeFile in Options then begin
                  WizardForm.ShowReadmeCheck.Visible := True;
                  ReadmeExecuteName := ChangeDirConstPChar(DestName);
                  Break;
                end
          end
          else begin
            with WizardForm do begin
              ChangeFinishedLabel (FixLabel(StrPas(SetupMessages[msgFinishedRestartLabel])));
              YesRadio.Visible := True;
              NoRadio.Visible := True;
            end;
          end;
          if not Assigned(InfoAfterText) then begin
            if not NeedsRestart and (shDisableFinishedPage in SetupHeader.Options) then
              goto NextStep;  { proceed to 'csTerminate' step }
            WizardForm.CurPage := wpFinished;
          end
          else
            WizardForm.CurPage := wpInfoAfter;
          Application.Restore;
          Update;
          WizardForm.CurPageChanged;
          WizardForm.CantShow := False;
          WizardForm.Show;
        end;
    csTerminate: Close;
  end;
end;

function ExitSetupMsgBox: Boolean;
begin
  Result := MsgBoxP(SetupMessages[msgExitSetupMessage], SetupMessages[msgExitSetupTitle],
    mbConfirmation, MB_YESNO or MB_DEFBUTTON2) = ID_YES;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if not InExecuteLoop then begin
    if (CurStep = csCopy) and Assigned(InstallForm) then begin
      CanClose := False;
      InstallForm.Close;
    end
    else
      if CurStep < csFinished then
        CanClose := ExitSetupMsgBox
      else
        CanClose := True;
  end
  else
    CanClose := False;
end;

procedure TMainForm.FormActivate(Sender: TObject);
const
  StartedSteps: Boolean = False;
begin
  { Start the 'steps' }
  if not StartedSteps then begin
    CurStep := csStart;
    Perform (WM_SETUPNEXTSTEP, 0, 0);
    StartedSteps := True;
  end;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Closing := True;
end;

procedure TMainForm.WMGetDlgCode (var Message: TWMGetDlgCode);
begin
  Message.Result := Message.Result or DLGC_WANTTAB;
end;

function EWP (Wnd: HWND; Param: Longint): BOOL;
{$IFDEF WIN32} stdcall; {$ELSE} export; {$ENDIF}
begin
  { Note: GetParent is not used here because the other windows are not
    actually child windows since they don't have WS_CHILD set. }
  {$IFDEF WIN32}
  if GetWindowLong(Wnd, GWL_HWNDPARENT) <> Param then
  {$ELSE}
  if GetWindowWord(Wnd, GWW_HWNDPARENT) <> Param then
  {$ENDIF}
    Result := True
  else begin
    Result := False;
    BringWindowToTop (Wnd);
  end;
end;
procedure TMainForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  { If, for some reason, the user doesn't have a mouse and the main form was
    activated, there would normally be no way to reactivate the wizard form.
    But this reactivates the form if the user hits a key on the keyboard }
  Key := 0;
  EnumWindows (@EWP, Handle);
end;

type
  PEWPRec = ^TEWPRec;
  TEWPRec = record
    Window: HWND;
    NotMinimized: Boolean;
  end;
function EWP2 (Wnd: HWND; Param: PEWPRec): BOOL;
{$IFDEF WIN32} stdcall; {$ELSE} export; {$ENDIF}
var
  Ctl: TWinControl;
begin
  Result := True;
  { Note: GetParent is not used here because the other windows are not
    actually child windows since they don't have WS_CHILD set. }
  {$IFDEF WIN32}
  if HWND(GetWindowLong(Wnd, GWL_HWNDPARENT)) = Param^.Window then begin
  {$ELSE}
  if GetWindowWord(Wnd, GWW_HWNDPARENT) = Param^.Window then begin
  {$ENDIF}
    Ctl := FindControl(Wnd);
    if Ctl is TSetupChildForm then
      with TSetupChildForm(Ctl) do
        Visible := Param^.NotMinimized and not CantShow;
  end;
end;
function TMainForm.MainWindowHook (var Message: TMessage): Boolean;
var
  E: TEWPRec;
  IsIcon: Boolean;
begin
  Result := False;
  case Message.Msg of
    WM_WINDOWPOSCHANGED: begin
        IsIcon := IsIconic(Application.Handle);
        if IsMinimized <> IsIcon then begin
          IsMinimized := IsIcon;
          E.Window := Handle;
          E.NotMinimized := not IsMinimized and IsWindowVisible(Handle);
          EnumWindows (@EWP2, Longint(@E));
        end;
      end;
  end;
end;
procedure TMainForm.AppOnRestore (Sender: TObject);
var
  E: TEWPRec;
begin
  IsMinimized := False;
  E.Window := Handle;
  E.NotMinimized := True;
  EnumWindows (@EWP2, Longint(@E));
end;

{ Work around problem in D2's declaration of the function }
{$IFDEF WIN32}
function NewAdjustTokenPrivileges(TokenHandle: THandle; DisableAllPrivileges: BOOL;
  const NewState: TTokenPrivileges; BufferLength: DWORD;
  PreviousState: PTokenPrivileges; ReturnLength: PDWORD): BOOL; stdcall;
  external advapi32 name 'AdjustTokenPrivileges';
{$ENDIF}

{$IFNDEF WIN32}
  {$I REBOOT.INC}
{$ENDIF}

procedure RestartComputer;
  procedure RestartErrorMessage;
  begin
    MessageBox (0, SetupMessages[msgErrorRestartingComputer],
      SetupMessages[msgErrorTitle], MB_OK or MB_ICONEXCLAMATION);
  end;
{$IFDEF WIN32}
var
  Token: THandle;
  TokenPriv: TTokenPrivileges;
const
  SE_SHUTDOWN_NAME = 'SeShutdownPrivilege';  { don't localize }
{$ELSE}
var
  Filename: String;
  F: File;
{$ENDIF}
begin
  {$IFDEF WIN32}
  if Win32Platform = VER_PLATFORM_WIN32_NT then begin
    if not OpenProcessToken(GetCurrentProcess, TOKEN_ADJUST_PRIVILEGES or TOKEN_QUERY,
       {$IFNDEF Delphi3orHigher} @Token {$ELSE} Token {$ENDIF}) then begin
      RestartErrorMessage;
      Exit;
    end;

    LookupPrivilegeValue (nil, SE_SHUTDOWN_NAME, TokenPriv.Privileges[0].Luid);

    TokenPriv.PrivilegeCount := 1;
    TokenPriv.Privileges[0].Attributes := SE_PRIVILEGE_ENABLED;

    NewAdjustTokenPrivileges (Token, False, TokenPriv, 0, nil, nil);

    { Cannot test the return value of AdjustTokenPrivileges. }
    if GetLastError <> ERROR_SUCCESS then begin
      RestartErrorMessage;
      Exit;
    end;
  end;
  if not ExitWindowsEx(EWX_REBOOT, 0) then
    RestartErrorMessage;
  {$ELSE}
  if not UsingWinNT then begin
    if not ExitWindows(EW_RESTARTWINDOWS, 0) then
      RestartErrorMessage;
  end
  else begin
    try
      Filename := GenerateUniqueName(GetTempDir, '.exe');
      AssignFile (F, Filename);
      FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (F, 1);
      try
        try
          BlockWrite (F, Reboot32EXE, SizeOf(Reboot32EXE));
        finally
          CloseFile (F);
        end;
      except
        DeleteFile (Filename);
        raise;
      end;
      try
        RestartReplace (Filename, '');
      except
        { ignore exceptions }
      end;
      if WinExec(StringAsPChar(Filename), SW_SHOW) < 32 then
        Abort;  {so that the error message gets displayed}
    except
      RestartErrorMessage;
    end;
  end;
  {$ENDIF}

  { If ExitWindows/ExitWindowsEx were successful, program execution halts here
    (at least on Win95) }
end;

procedure DeinitSetup;
var
  I: Integer;
begin
  for I := 0 to DeleteFilesAfterInstallList.Count-1 do
    DeleteFile (DeleteFilesAfterInstallList[I]);
  DeleteFilesAfterInstallList.Clear;
  for I := 0 to DeleteDirsAfterInstallList.Count-1 do
    {$IFNDEF WIN32}
    try
      RmDir (DeleteDirsAfterInstallList[I]);
    except
    end;
    {$ELSE}
    RemoveDirectory (PChar(DeleteDirsAfterInstallList[I]));
    {$ENDIF}
  DeleteDirsAfterInstallList.Clear;

  if TempInstallDir <> '' then
    DelTree (TempInstallDir, True, True, True, nil, nil);

  if RestartSystem then
    RestartComputer;
end;

procedure ExitProc; far;
begin
  WizardImage.Free;
  WizardImage := nil;
end;

initialization
  IsNT := UsingWinNT;
  WinVersion := Swap({$IFDEF WIN32} GetVersion {$ELSE} ThunkGetVersion {$ENDIF});
  NewGUI := WinVersion >= $400;
  DirEntries := TList.Create;
  FileEntries := TList.Create;
  FileLocationEntries := TList.Create;
  IconEntries := TList.Create;
  IniEntries := TList.Create;
  RegistryEntries := TList.Create;
  InstallDeleteEntries := TList.Create;
  UninstallDeleteEntries := TList.Create;
  RunEntries := TList.Create;
  UninstallRunEntries := TList.Create;
  DeleteFilesAfterInstallList := TStringList.Create;
  DeleteDirsAfterInstallList := TStringList.Create;
{$IFNDEF WIN32}
  AddExitProc (ExitProc);
{$ELSE}
finalization
  ExitProc;
{$ENDIF}
end.
