program Uninst;

{
  Inno Setup
  Copyright (C) 1998-99 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  Uninstaller
}

uses
  WinProcs,
  WinTypes,
  SysUtils,
  Messages,
  CmnFunc2 in 'CMNFUNC2.PAS',
  Undo in 'UNDO.PAS',
  Msgs in 'MSGS.PAS',
  MsgIDs in 'MSGIDS.PAS',
  InstFunc in 'INSTFUNC.PAS';

{$IFDEF WIN32}
  {$R *.R32}
{$ELSE}
  {$R *.R16}
{$ENDIF}

function MessageBoxFmt1 (const ID: TSetupMessageID; const Arg1: String;
  const Title: PChar; const Flags: UINT): Integer;
{$IFNDEF WIN32}
var
  Msg: array[0..255] of Char;
{$ENDIF}
begin
  Result := MessageBox(0,
    {$IFDEF WIN32} PChar(FmtSetupMessage1(ID, Arg1)) {$ELSE} StrPCopy(Msg, FmtSetupMessage1(ID, Arg1)) {$ENDIF},
    Title, Flags);
end;

procedure ShowExceptionMsg (const Msg: String);
var
  {$IFDEF WIN32}
  M: String;
  {$ELSE}
  M: array[0..256] of Char;  { not 255, since it may add a period to the end }
  {$ENDIF}
begin
  {$IFDEF WIN32}
  M := Msg;
  if (M <> '') and (M[Length(M)] > '.') then M := M + '.';
  {$ELSE}
  StrPCopy (M, Msg);
  if (Msg <> '') and (Msg[Length(Msg)] > '.') then
    lstrcat (M, '.');  { use Windows' cat for minimum code size }
  {$ENDIF}
  MessageBox (0, {$IFDEF WIN32} PChar( {$ENDIF} M {$IFDEF WIN32} ) {$ENDIF},
    SetupMessages[msgErrorTitle], MB_OK or MB_ICONSTOP);
    { ^ referencing SetupMessages[msgErrorTitle] will still work even if
      the messages aren't loaded yet, since MessageBox displays 'Error'
      as the caption if the lpCaption parameter is nil. }
end;

procedure RaiseLastError (const Msg: TSetupMessageID);
{$IFDEF WIN32}
var
  ErrorCode: DWORD;
{$ENDIF}
begin
  {$IFDEF WIN32}
  ErrorCode := GetLastError;
  raise Exception.Create(FmtSetupMessage(msgLastErrorMessage,
    [SetupMessages[Msg], IntToStr(ErrorCode), SysErrorMessage(ErrorCode)]));
  {$ELSE}
  raise Exception.Create(StrPas(SetupMessages[Msg]));
  {$ENDIF}
end;

const
  SecondPhaseWndName = '_InnoSetupUninstallerWindow_';  { don't localize }

function EnumThreadWindowsProc (Wnd: HWND; LParam: Longint): BOOL;
{$IFDEF WIN32} stdcall; {$ELSE} export; {$ENDIF}
begin
  Result := False;
  {$IFDEF WIN32} SetForegroundWindow {$ELSE} SetActiveWindow {$ENDIF} (Wnd);
end;

function Exec (const Filename: String;
  const Parms: {$IFDEF WIN32} String {$ELSE} PChar {$ENDIF}): THandle;
var
  {$IFDEF WIN32}
  CmdLine: String;
  StartupInfo: TStartupInfo;
  ProcessInfo: TProcessInformation;
  {$ELSE}
  CmdLine: array[0..1023] of Char;
  H: THandle;
  {$ENDIF}
begin
  {$IFDEF WIN32}
  CmdLine := AddQuotes(Filename) + ' ' + Parms;
  {$ELSE}
  StrPCopy (CmdLine, Filename);
  StrLCat (CmdLine, ' ', SizeOf(CmdLine)-1);
  StrLCat (CmdLine, Parms, SizeOf(CmdLine)-1);
  {$ENDIF}

  {$IFDEF WIN32}
  FillChar (StartupInfo, SizeOf(StartupInfo), 0);
  StartupInfo.cb := SizeOf(StartupInfo);
  if not CreateProcess(nil, PChar(CmdLine), nil, nil, False, 0, nil, nil,
     StartupInfo, ProcessInfo) then
    RaiseLastError (msgLdrCannotExecTemp);
  CloseHandle (ProcessInfo.hThread);
  Result := ProcessInfo.hProcess;
  {$ELSE}
  Result := WinExec(CmdLine, SW_SHOW);
  if Result < 32 then
    RaiseLastError (msgLdrCannotExecTemp);
  {$ENDIF}
end;

function GenerateUniqueExeName (Path: String; var Filename: String): Boolean;
{ Returns True if it overwrote an existing file. }
  function IntToBase32 (Number: Longint): String;
  const
    Table: array[0..31] of Char = '0123456789ABCDEFGHIJKLMNOPQRSTUV';
  var
    I: Integer;
  begin
    Result := '';
    for I := 0 to 4 do begin
      Insert (Table[Number and 31], Result, 1);
      Number := Number shr 5;
    end;
  end;
var
  Rand, RandOrig: Longint;
  F: {$IFDEF WIN32} THandle {$ELSE} File {$ENDIF};
  Success: Boolean;
begin
  Path := AddBackslash(Path);
  RandOrig := $123456;
  Rand := RandOrig;
  Success := False;
  Result := False;
  repeat
    Inc (Rand);
    if Rand > $1FFFFFF then Rand := 0;
    if Rand = RandOrig then
      { practically impossible to go through 33 million possibilities,
        but check "just in case"... }
      raise Exception.Create(FmtSetupMessage1(msgErrorTooManyFilesInDir,
        RemoveBackslashUnlessRoot(Path)));
    { Generate a random name }
    Filename := Path + '_iu' + IntToBase32(Rand) + '.tmp';
    if DirExists(Filename) then Continue;
    Success := True;
    Result := FileExists(Filename);
    if Result then begin
      {$IFDEF WIN32}
      F := CreateFile(PChar(Filename), GENERIC_READ or GENERIC_WRITE, 0,
        nil, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
      Success := F <> INVALID_HANDLE_VALUE;
      if Success then CloseHandle (F);
      {$ELSE}
      try
        AssignFile (F, Filename);
        FileMode := fmOpenReadWrite or fmShareExclusive;
        Rewrite (F, 1);
        CloseFile (F);
      except
        Success := False;
      end;
      {$ENDIF}
    end;
  until Success;
end;

function ProcessMsgs: Boolean;
var
  Msg: TMsg;
begin
  Result := False;
  while PeekMessage(Msg, 0, 0, 0, PM_REMOVE) do begin
    if Msg.Message = WM_QUIT then begin
      Result := True;
      Break;
    end;
    TranslateMessage (Msg);
    DispatchMessage (Msg);
  end;
end;

const
  WM_KillFirstPhase = WM_USER + 333;
var
  OldWindowProc: Pointer;

{$IFDEF WIN32}
function FirstPhaseWindowProc (Wnd: HWND; Msg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT; stdcall;
{$ELSE}
function FirstPhaseWindowProc (Wnd: HWND; Msg: Word; wParam: Word;
  lParam: Longint): Longint; export;
{$ENDIF}
begin
  if Msg = WM_KillFirstPhase then begin
    PostQuitMessage (0);
    Result := 0;
  end
  else
    Result := CallWindowProc(OldWindowProc, Wnd, Msg, wParam, lParam);
end;

{$IFDEF WIN32}
procedure DelayDeleteFile (const Filename: String);
var
  I: Integer;
begin
  for I := 0 to 11 do begin
    if I <> 0 then Sleep (250);
    if WinProcs.DeleteFile(PChar(Filename)) or
       (GetLastError = ERROR_FILE_NOT_FOUND) or
       (GetLastError = ERROR_PATH_NOT_FOUND) then
      Break;
  end;
end;
{$ENDIF}

const
  ParamName = {$IFDEF WIN32} '/EUFT32A' {$ELSE} '/EUFT16A' {$ENDIF};
  SelfReadMode =
    {$IFDEF WIN32}
    fmOpenRead or fmShareDenyWrite
    {$ELSE}
    fmOpenRead
    { ^ On Windows 3.1 w/ SHARE and without VSHARE.386, a sharing violation
      message can appear if an EXE opens itself with the "fmShareDenyWrite"
      mode. So only use fmOpenRead on Win16 }
    {$ENDIF};

var
  TempFile, UninstExeFile, UninstDataFile, UninstMsgFile: String;
  UndoChanges: TUndoChanges {$IFDEF WIN32} = nil {$ENDIF};
  Title {$IFNDEF WIN32} , ExceptMsg {$ENDIF}: array[0..255] of Char;
  SecondPhase, RemovedAll: Boolean;
  FirstPhaseWnd, SecondPhaseWnd: HWND;
  ProcessHandle: THandle;
  F: File;
  {$IFDEF WIN32}
  Quit: Boolean = False;
  {$ELSE}
  CmdLine: array[0..2047] of Char;
  {$ENDIF}
const
  RemovedMsgs: array[Boolean] of TSetupMessageID =
    (msgUninstalledMost, msgUninstalledAll);

procedure DeleteUninstallDataFiles; far;
{$IFDEF WIN32}
var
  ProcessID: DWORD;
  Process: THandle;
{$ENDIF}
begin
  DeleteFile (UninstDataFile);
  DeleteFile (UninstMsgFile);
  {$IFDEF WIN32}
  GetWindowThreadProcessId (FirstPhaseWnd, @ProcessID);
  Process := OpenProcess(STANDARD_RIGHTS_REQUIRED or SYNCHRONIZE, False,
    ProcessID);
  SendMessage (FirstPhaseWnd, WM_KillFirstPhase, 0, 0);
  WaitForSingleObject (Process, INFINITE);
  CloseHandle (Process);
  Sleep (0);  { give first phase a chance to terminate completely }
  DelayDeleteFile (UninstExeFile);
  {$ELSE}
  SendMessage (FirstPhaseWnd, WM_KillFirstPhase, 0, 0);
  ProcessMsgs;  { give first phase a chance to terminate completely }
  DeleteFile (UninstExeFile);
  {$ENDIF}
end;

begin
  try
    {$IFNDEF WIN32}
    try ChDir (GetWinDir); except end;
    {$ELSE}
    SetCurrentDirectory (PChar(GetWinDir));
    {$ENDIF}

    if ParamCount = 0 then begin
      SecondPhase := False;
      UninstExeFile := ParamStr(0);
    end
    else
    if ParamCount = 3 then begin
      if ParamStr(1) <> ParamName then Exit;
      SecondPhase := True;
      UninstExeFile := ParamStr(2);
      FirstPhaseWnd := StrToInt(ParamStr(3));
    end
    else
      Exit;

    AssignFile (F, ParamStr(0));
    FileMode := SelfReadMode;
    Reset (F, 1);
    try
      UninstMsgFile := ChangeFileExt(UninstExeFile, '.msg');
      LoadSetupMessages (UninstMsgFile, 0, fmOpenRead or fmShareDenyWrite);
      UninstDataFile := ChangeFileExt(UninstExeFile, '.dat');

      { Search for previous uninstaller instances, and if one is found,
        activate its window. }
      SecondPhaseWnd := FindWindow(nil, SecondPhaseWndName);
      if SecondPhaseWnd <> 0 then begin
        {$IFDEF WIN32}
        EnumThreadWindows (GetWindowThreadProcessId(SecondPhaseWnd, nil),
          @EnumThreadWindowsProc, 0);
        {$ELSE}
        EnumTaskWindows (GetWindowTask(SecondPhaseWnd),
          @EnumThreadWindowsProc, 0);
        {$ENDIF}
        Exit;
      end;

      if not SecondPhase then begin
        { Verify that uninstall data file exists }
        if not FileExists(UninstDataFile) then begin
          MessageBoxFmt1 (msgUninstallNotFound, UninstDataFile,
            SetupMessages[msgUninstallAppTitle], MB_ICONSTOP or MB_OK);
          Exit;
        end;

        { Copy self to TEMP directory with a name like _iu14D2N.tmp. The
          actual uninstallation process must be done from somewhere outside
          the application directory since EXE's can't delete themselves while
          they are running. }
        if not GenerateUniqueExeName(GetTempDir, TempFile)
           {$IFNDEF WIN32} and (UsingWinNT or (Lo(ThunkGetVersion) >= 4)) {$ENDIF} then
           { ^ Windows 3.1 does NOT support "NUL" entries in wininit.ini }
          try
            RestartReplace (TempFile, '');
          except
            { ignore exceptions }
          end;
        if not FileCopy(UninstExeFile, TempFile, False, SelfReadMode) then begin
          RaiseLastError (msgLdrCannotCreateTemp);
          Exit;
        end;

        { Create FirstPhaseWnd. FirstPhaseWnd waits for a WM_KillFirstPhase
          message from the second phase process, and terminates itself in
          response. The reason the first phase doesn't just terminate
          immediately is because the Control Panel Add/Remove applet refreshes
          its list as soon as the program terminates. So it waits until the
          uninstallation is complete before terminating. }
        FirstPhaseWnd := CreateWindowEx(0, 'STATIC', '', 0, 0, 0, 0, 0,
          HWND_DESKTOP, 0, HInstance, nil);  { do not localize }
        Longint(OldWindowProc) := SetWindowLong(FirstPhaseWnd, GWL_WNDPROC,
          Longint(@FirstPhaseWindowProc));
        try
          { Execute the copy of itself ("second phase") }
          {$IFDEF WIN32}
          ProcessHandle := Exec(TempFile, Format(ParamName + ' %s $%x',
            [AddQuotes(ParamStr(0)), FirstPhaseWnd]));
          {$ELSE}
          StrLFmt (CmdLine, SizeOf(CmdLine)-1, ParamName + ' %s $%x',
            [ParamStr(0), FirstPhaseWnd]);
          ProcessHandle := Exec(TempFile, CmdLine);
          {$ENDIF}

          { Wait till the second phase process unexpectedly dies or is ready
            for the first phase to terminate. }
          {$IFDEF WIN32}
          repeat until ProcessMsgs or (MsgWaitForMultipleObjects(1,
            ProcessHandle, False, INFINITE, QS_ALLINPUT) <> WAIT_OBJECT_0+1);
          CloseHandle (ProcessHandle);
          {$ELSE}
          repeat until (GetModuleUsage(ProcessHandle) = 0) or ProcessMsgs;
          {$ENDIF}
        finally
          DestroyWindow (FirstPhaseWnd);
        end;
      end
      else begin
        { Create a window with the name SecondPhaseWndName. In order to prevent
          multiple instances, the uninstaller searches for this window at
          startup and exits if it finds it already exists. }
        SecondPhaseWnd := CreateWindowEx(0, 'STATIC', SecondPhaseWndName,
          0, 0, 0, 0, 0, HWND_DESKTOP, 0, HInstance, nil);  { do not localize }
        try
          UndoChanges := TUndoChanges.Create;
          UndoChanges.Load (UninstDataFile);
          StrPLCopy (Title, FmtSetupMessage(msgUninstallAppFullTitle,
            UndoChanges.AppName), SizeOf(Title)-1);

          {$IFDEF WIN32}
          if UndoChanges.AdminInstalled and not IsAdminLoggedOn then begin
            MessageBox (0, SetupMessages[msgOnlyAdminCanUninstall], Title,
              MB_OK or MB_ICONEXCLAMATION);
            Exit;
          end;
          {$ENDIF}

          if MessageBoxFmt1(msgConfirmUninstall, UndoChanges.AppName, Title,
             MB_ICONQUESTION or MB_YESNO or MB_DEFBUTTON2) <> ID_YES then
            Exit;

          { Start the actual uninstall process }
          RemovedAll := UndoChanges.PerformUndo(True, DeleteUninstallDataFiles);

          MessageBoxFmt1 (RemovedMsgs[RemovedAll], UndoChanges.AppName,
            Title, MB_ICONINFORMATION or MB_OK
              {$IFDEF WIN32} or MB_SETFOREGROUND {$ENDIF});
        finally
          UndoChanges.Free;
          DestroyWindow (SecondPhaseWnd);
        end;
      end;
    finally
      CloseFile (F);
    end;
  except
    on E: Exception do
      ShowExceptionMsg (E.Message);
  end;
end.
