unit Wizard;

{
  Inno Setup
  Copyright (C) 1998-99 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  Wizard form
}

interface

uses
  WinTypes, WinProcs, SysUtils, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, NewBevel, NewFCtrl, ExtCtrls,
  ChildFrm;

type
  TWizardPage = (wpWelcome, wpPassword, wpLicense, wpInfoBefore, wpSelectDir,
    wpSelectProgramGroup, wpReady, wpInfoAfter, wpFinished);

  TWizardForm = class(TSetupChildForm)
    CancelButton: TButton;
    NextButton: TButton;
    BackButton: TButton;
    Notebook1: TNotebook;
    PicturePaintBox: TPaintBox;
    Bevel: TBevel;
    ClickLabel: TLabel;
    Notebook2: TNotebook;
    WelcomeBox: TPaintBox;
    SelectFolderLabel: TLabel;
    DiskSpaceLabel: TLabel;
    DirEdit: TEdit;
    DirList: TNewDirectoryListBox;
    DriveList: TNewDriveComboBox;
    IconsLabel: TLabel;
    GroupEdit: TEdit;
    NoIconsCheck: TCheckBox;
    GroupList: TListBox;
    LicenseLabel1: TLabel;
    LicenseMemo: TMemo;
    LicenseLabel2: TLabel;
    LicenseIcon: TPaintBox;
    ReadmeBeforeIcon: TPaintBox;
    InfoBeforeLabel: TLabel;
    InfoBeforeMemo: TMemo;
    ReadmeAfterIcon: TPaintBox;
    InfoAfterLabel: TLabel;
    InfoAfterMemo: TMemo;
    FinishedLabel: TLabel;
    ShowReadmeCheck: TCheckBox;
    InfoBeforeClickLabel: TLabel;
    InfoAfterClickLabel: TLabel;
    YesRadio: TRadioButton;
    NoRadio: TRadioButton;
    ReadyLabel: TLabel;
    PasswordLabel: TLabel;
    PasswordEdit: TEdit;
    PasswordEditLabel: TLabel;
    procedure NextButtonClick(Sender: TObject);
    procedure BackButtonClick(Sender: TObject);
    procedure CancelButtonClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure DirListChange(Sender: TObject);
    procedure GroupListClick(Sender: TObject);
    procedure NoIconsCheckClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure PicturePaintBoxPaint(Sender: TObject);
    procedure GroupEditChange(Sender: TObject);
    procedure WelcomeBoxPaint(Sender: TObject);
    procedure SetupIconPaint(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    DisableCloseQuery: Boolean;
    ShowInitDone: Boolean;
    procedure ReadGroupList;
    procedure WMSysCommand (var Message: TWMSysCommand); message WM_SYSCOMMAND;
  public
    { Public declarations }
    CurPage: TWizardPage;
    OrigFinishedLabelWidth: Integer;
    procedure ChangeFinishedLabel (const S: String);
    procedure CurPageChanged;
    function SkipCurPage: Boolean;
  end;

var
  WizardForm: TWizardForm;

function FixLabel (const S: String): String;

implementation

uses
  {$IFDEF WIN32} ShlObj, {$ENDIF}
  Msgs, MsgIDs, Main, Install, Struct,
  CmnFunc, CmnFunc2,
  DDEInt, zlib;

{$R *.DFM}

const
  NotebookPages: array[TWizardPage, 0..1] of Integer =
    ((0, 0), (0, 1), (1, -1), (2, -1), (0, 2), (0, 3), (0, 4), (3, -1), (0, 5));
  PageCaptions: array[TWizardPage] of TSetupMessageID =
    (msgWizardWelcome, MsgWizardPassword, msgWizardLicense, msgWizardInfoBefore,
     msgWizardSelectDir, msgWizardSelectProgramGroup, msgWizardReady,
     msgWizardInfoAfter, msgWizardFinished);

var
  MinimumSpaceStr: {$IFDEF WIN32} String; {$ELSE} String[15]; {$ENDIF}

procedure CalcMinimumSpaceStr;
var
  MinMB, MinMBFrac, MinMBDec: Longint;
begin
  MinMB := MinimumSpace div 1048576;
  MinMBFrac := (MinimumSpace mod 1048576) * 10;
  MinMBDec := MinMBFrac div 1048576;
  if MinMBFrac mod 1048576 <> 0 then  { round up }
    Inc (MinMBDec);
  if MinMBDec = 10 then begin
    Inc (MinMB);
    MinMBDec := 0;
  end;
  MinimumSpaceStr := Format('%d.%d', [MinMB, MinMBDec]);
end;

procedure FixLabelBuf (Buffer: PChar; MaxLen: Cardinal; S: PChar);
var
  Buf: PChar;

  procedure Append (Str: PChar);
  var
    L: Cardinal;
  begin
    L := StrLen(Str);
    if L > MaxLen then L := MaxLen;
    Buf := StrEnd(StrLCopy(Buf, Str, L));
    Dec (MaxLen, L);
  end;
var
  P: PChar;
  L: Cardinal;
  MB: array[0..15] of Char;
begin
  Buf := Buffer;
  Buf[0] := #0;
  while MaxLen > 0 do begin
    P := StrScan(S, '[');
    if P = nil then begin
      StrLCopy (Buf, S, MaxLen);
      Break;
    end;
    if P <> S then begin
      L := P - S;
      if L > MaxLen then L := MaxLen;
      Buf := StrEnd(StrLCopy(Buf, S, L));
      Dec (MaxLen, L);
      if MaxLen = 0 then Break;
      S := P;
    end;
    if StrLComp(P, '[name]', 6) = 0 then begin
      Append (SetupHeader.AppName);
      Inc (S, 6);
    end
    else
    if StrLComp(P, '[name/ver]', 10) = 0 then begin
      Append (SetupHeader.AppVerName);
      Inc (S, 10);
    end
    else
    if StrLComp(P, '[mb]', 4) = 0 then begin
      if MinimumSpaceStr = '' then
        CalcMinimumSpaceStr;
      StrPLCopy (MB, MinimumSpaceStr, SizeOf(MB)-1);
      Append (MB);
      Inc (S, 4);
    end
    else begin
      Append ('[');
      Inc (S);
    end;
  end;
end;

function FixLabel (const S: String): String;
begin
  Result := S;
  {don't localize these}
  StringChange (Result, '[name]', StrPas(SetupHeader.AppName));
  StringChange (Result, '[name/ver]', StrPas(SetupHeader.AppVerName));
  if MinimumSpaceStr = '' then
    CalcMinimumSpaceStr;
  StringChange (Result, '[mb]', MinimumSpaceStr);
end;

procedure TWizardForm.FormCreate(Sender: TObject);
var
  X, W1, W2, W3: Integer;
  S: String;
  SystemMenu: HMENU;
const
  idDirectories: array[Boolean] of TSetupMessageID =
    (msgDirectoryOld, msgDirectoryNew);
  idProgramManagers: array[Boolean] of TSetupMessageID =
    (msgProgramManagerOld, msgProgramManagerNew);
  {}WizardButtonWidth = 75;
begin
  { Give it a minimize button if main window isn't visible }
  if not(shWindowVisible in SetupHeader.Options) then begin
    BorderIcons := BorderIcons + [biMinimize];
    BorderStyle := bsSingle;
  end;

  { Position the buttons }
  W1 := MulDiv(WizardButtonWidth, Screen.PixelsPerInch, 96);  { width of each button }
  { ^ In small fonts PixelsPerInch is 96, so W1 = WizardButtonWidth.
    Large fonts have a higher PixelsPerInch, so W1 is scaled
    accordingly. }
  W2 := MulDiv(8, Screen.PixelsPerInch, 96);   { space between edge of form and Cancel button }
  W3 := MulDiv(10, Screen.PixelsPerInch, 96);  { space between Next and Cancel button }
  BackButton.Width := W1;
  NextButton.Width := W1;
  CancelButton.Width := W1;
  X := ClientWidth - W2 - W1;
  CancelButton.Left := X;
  Dec (X, W3);
  Dec (X, W1);
  NextButton.Left := X;
  Dec (X, W1);
  BackButton.Left := X;

  DirEdit.MaxLength := {$IFNDEF WIN32} 63 {$ELSE} 255 {$ENDIF};

  { Assign captions }
  ClickLabel.SetTextBuf (SetupMessages[msgClickNext]);
  PasswordLabel.SetTextBuf (SetupMessages[msgPasswordLabel]);
  PasswordEditLabel.SetTextBuf (SetupMessages[msgPasswordEditLabel]);
  LicenseLabel1.Caption := FixLabel(StrPas(SetupMessages[msgLicenseLabel1]));
  LicenseLabel2.Caption := FixLabel(StrPas(SetupMessages[msgLicenseLabel2]));
  InfoBeforeLabel.SetTextBuf (SetupMessages[msgInfoBeforeLabel]);
  InfoBeforeClickLabel.SetTextBuf (SetupMessages[msgInfoBeforeClickLabel]);
  DiskSpaceLabel.Caption := FixLabel(StrPas(SetupMessages[msgDiskSpaceMBLabel]));
  NoIconsCheck.SetTextBuf (SetupMessages[msgNoIconsCheck]);
  InfoAfterLabel.SetTextBuf (SetupMessages[msgInfoAfterLabel]);
  InfoAfterClickLabel.SetTextBuf (SetupMessages[msgInfoAfterClickLabel]);
  S := FixLabel(StrPas(SetupMessages[msgReadyLabel1])) + SNewLine2;
  if not(shDisableDirPage in SetupHeader.Options) or
     (HasIcons and not(shDisableProgramGroupPage in SetupHeader.Options)) then
    S := S + StrPas(SetupMessages[msgReadyLabel2a])
  else
    S := S + StrPas(SetupMessages[msgReadyLabel2b]);
  ReadyLabel.Caption := S;
  if HasIcons then
    S := FixLabel(StrPas(SetupMessages[msgFinishedLabel]))
  else
    S := FixLabel(StrPas(SetupMessages[msgFinishedLabelNoIcons]));
  OrigFinishedLabelWidth := FinishedLabel.Width;
  ChangeFinishedLabel (S + SNewLine2 + StrPas(SetupMessages[msgClickFinish]));
  ShowReadmeCheck.SetTextBuf (SetupMessages[msgShowReadmeCheck]);
  YesRadio.SetTextBuf (SetupMessages[msgYesRadio]);
  NoRadio.SetTextBuf (SetupMessages[msgNoRadio]);
  SelectFolderLabel.Caption := FixLabel(FmtSetupMessage1(msgSelectFolderLabel,
    StrPas(SetupMessages[idDirectories[NewGUI]])));
  IconsLabel.Caption := FmtSetupMessage1(msgIconsLabel,
    StrPas(SetupMessages[idProgramManagers[NewGUI]]));

  if Assigned(LicenseText) then
    LicenseMemo.SetTextBuf (LicenseText);
  if Assigned(InfoBeforeText) then
    InfoBeforeMemo.SetTextBuf (InfoBeforeText);
  if Assigned(InfoAfterText) then
    InfoAfterMemo.SetTextBuf (InfoAfterText);

  { Center the form inside the MainForm's client area }
  MainForm.CenterForm (Self);

  { Append an 'About Setup' item to the wizard form also }
  SystemMenu := GetSystemMenu(Handle, False);
  AppendMenu (SystemMenu, MF_SEPARATOR, 0, nil);
  AppendMenu (SystemMenu, MF_STRING, 9999, SetupMessages[msgAboutSetupMenuItem]);
end;

procedure TWizardForm.FormDestroy(Sender: TObject);
begin
  { empty }
end;

procedure TWizardForm.ChangeFinishedLabel (const S: String);
var
  Y: Integer;
begin
  FinishedLabel.AutoSize := False;
  FinishedLabel.Width := OrigFinishedLabelWidth;
  FinishedLabel.Caption := S + SNewLine;
  FinishedLabel.AutoSize := True;
  Y := FinishedLabel.Top + FinishedLabel.Height;
  ShowReadmeCheck.Top := Y;
  YesRadio.Top := Y;
  NoRadio.Top := Y + MulDiv(24, Screen.PixelsPerInch, 96);
end;

procedure TWizardForm.CurPageChanged;
{ Call this whenever the current page is changed }
var
  V: Boolean;
label 1;
begin
  Notebook1.PageIndex := NotebookPages[CurPage, 0];
  if NotebookPages[CurPage, 1] <> -1 then
    Notebook2.PageIndex := NotebookPages[CurPage, 1];

  { Set button visibility and captions }
  V := not(CurPage in [Low(TWizardPage), wpInfoAfter, wpFinished]);
  if Assigned(InfoAfterText) and (CurPage = wpFinished) then
    V := True;
  BackButton.Visible := V;
  CancelButton.Visible := CurPage < wpInfoAfter;
  BackButton.SetTextBuf (SetupMessages[msgButtonBack]);
  case CurPage of
    wpLicense: begin
        NextButton.SetTextBuf (SetupMessages[msgButtonYes]);
        CancelButton.SetTextBuf (SetupMessages[msgButtonNo]);
      end;
    wpReady: begin
        NextButton.SetTextBuf (SetupMessages[msgButtonInstall]);
        CancelButton.SetTextBuf (SetupMessages[msgButtonCancel]);
      end;
    wpFinished: begin
        NextButton.SetTextBuf (SetupMessages[msgButtonFinish]);
        CancelButton.SetTextBuf (SetupMessages[msgButtonCancel]);
      end;
  else
  1:NextButton.SetTextBuf (SetupMessages[msgButtonNext]);
    CancelButton.SetTextBuf (SetupMessages[msgButtonCancel]);
  end;

  { Set the Caption to match the current page's title }
  SetTextBuf (SetupMessages[PageCaptions[CurPage]]);

  { Hide ClickLabel if it's on the Welcome, Ready to Install, or Setup
    Completed wizard pages }
  ClickLabel.Visible := not(CurPage in [wpWelcome, wpReady, wpFinished]);

  { Don't make the Next button "default" (i.e. selectable with the Enter key)
    when on the License page }
  NextButton.Default := CurPage <> wpLicense;
  { Adjust focus }
  case CurPage of
    wpWelcome: ActiveControl := NextButton;
    wpPassword: ActiveControl := PasswordEdit;
    wpLicense: ActiveControl := LicenseMemo;
    wpInfoBefore: ActiveControl := InfoBeforeMemo;
    wpSelectDir: ActiveControl := DirEdit;
    wpSelectProgramGroup: begin
        if not NoIconsCheck.Checked then
          ActiveControl := GroupEdit
        else
          ActiveControl := NoIconsCheck;
      end;
    wpReady: ActiveControl := NextButton;
    wpInfoAfter: ActiveControl := InfoAfterMemo;
    wpFinished:
      if ShowReadmeCheck.Visible then
        ActiveControl := ShowReadmeCheck
      else
      if YesRadio.Visible and YesRadio.Checked then
        ActiveControl := YesRadio
      else
      if NoRadio.Visible and NoRadio.Checked then
        ActiveControl := NoRadio
      else
        ActiveControl := NextButton;
  end;
end;

function TWizardForm.SkipCurPage: Boolean;
begin
  Result :=
    ((CurPage = wpPassword) and not NeedPassword) or
    ((CurPage = wpLicense) and not Assigned(LicenseText)) or
    ((CurPage = wpInfoBefore) and not Assigned(InfoBeforeText)) or
    ((CurPage = wpSelectDir) and ((shDisableDirPage in SetupHeader.Options) or not(shCreateAppDir in SetupHeader.Options))) or
    ((CurPage = wpSelectProgramGroup) and ((shDisableProgramGroupPage in SetupHeader.Options) or not HasIcons));
end;

procedure TWizardForm.NextButtonClick(Sender: TObject);
  function CheckPassword: Boolean;
  var
    S: String;
    SaveCursor: HCURSOR;
    {$IFNDEF WIN32}
    EndTime: Longint;
    {$ENDIF}
  begin
    S := PasswordEdit.Text;
    Result := GetCRC32(S[1], Length(S)) = SetupHeader.Password;
    if Result then
      NeedPassword := False
    else begin
      { Delay for 750 ms when an incorrect password is entered to
        discourage brute-force attempts }
      SaveCursor := GetCursor;
      SetCursor (LoadCursor(0, IDC_WAIT));
      {$IFDEF WIN32}
      Sleep (750);
      {$ELSE}
      EndTime := GetTickCount + 750;
      while GetTickCount < EndTime do Yield;
      {$ENDIF}
      SetCursor (SaveCursor);
      MsgBoxP (SetupMessages[msgIncorrectPassword], nil, mbError, MB_OK);
      PasswordEdit.Text := '';
      PasswordEdit.SetFocus;
    end;
  end;
  function GetTotalFreeSpace (const Drive: Char; var TotalSpace: Longint): Boolean;
  var
    {$IFDEF WIN32}
    SectorsPerCluster, BytesPerSector, FreeClusters, TotalClusters: DWORD;
    {$ELSE}
    SectorsPerCluster, BytesPerSector, FreeClusters, TotalClusters: Word;
    DriveNum: Byte;
    {$ENDIF}
  begin
    {$IFDEF WIN32}
    Result := GetDiskFreeSpace(PChar(Drive + ':\'), SectorsPerCluster, BytesPerSector, FreeClusters, TotalClusters);
    {$ELSE}
    DriveNum := Ord(UpCase(Drive)) - 64;
    asm
       mov  ah, 36h
       mov  dl, DriveNum
       int  21h
       mov  Result, False
       cmp  ax, 0FFFFh
       je   @done
       mov  Result, True
       mov  SectorsPerCluster, ax
       mov  BytesPerSector, cx
       mov  FreeClusters, bx
       mov  TotalClusters, dx
      @done:
    end;
    {$ENDIF}
    if Result then
      {$IFDEF WIN32}
      TotalSpace := BytesPerSector * SectorsPerCluster * FreeClusters;
      {$ELSE}
      { ensure a 32-bit multiply }
      TotalSpace := Longint(BytesPerSector) * Longint(SectorsPerCluster) * Longint(FreeClusters);
      {$ENDIF}
  end;
  {$IFNDEF WIN32}
  function BadPath (Path: String): Boolean;
  { Note: Path is only the part of the path after the drive letter, colon, and
    first backslash }
  var
    I, J, K: Integer;
  begin
    Result := False;
    while Path <> '' do begin
      I := Pos('\', Path);
      if I = 0 then I := Length(Path)+1;
      J := Pos('.', Path);
      if J > I then J := 0;
      K := I-1;
      if J <> 0 then K := J-1;
      if (K > 8) or ((J <> 0) and ((I-J-1 < 1) or (I-J-1 > 3))) then begin
        Result := True;
        Break;
      end;
      Path := Copy(Path, I+1, Maxint);
    end;
  end;
  {$ENDIF}
  function SpaceString (const S: String): String;
  var
    I: Integer;
  begin
    Result := '';
    for I := 1 to Length(S) do begin
      if S[I] = ' ' then Continue;
      if Result <> '' then Result := Result + ' ';
      Result := Result + S[I];
    end;
  end;
  function CheckSelectDirPage: Boolean;
  const
    {$IFDEF WIN32}
    BadDirChars3 = '/:*?"<>|,()[]';
    BadDirChars4 = '/:*?"<>|';
    {$ELSE}
    BadDirChars = ' /:*?"<>|,;[]=+%^';  { Note: the space is intentional }
    {$ENDIF}
  var
    T: String;
    I: Integer;
    TotalFreeSpace: Longint;
    {$IFDEF WIN32}
    BadDirChars: String;
    {$ENDIF}
  begin
    Result := False;

    { Remove any leading or trailing spaces }
    DirEdit.Text := Trim(DirEdit.Text);

    T := DirEdit.Text;  { reduce calls to GetText }

    { Check for UNC pathname }
    if Pos('\\', T) <> 0 then begin
      MsgBoxP (SetupMessages[msgToUNCPathname], nil, mbError, MB_OK);
      Exit;
    end;

    { Check if is at least 4 chars long and it has a colon and backslash }
    if not(shAllowRootDirectory in SetupHeader.Options) then
      I := 4
    else
      I := 3;
    if (Length(T) < I) or (T[2] <> ':') or (T[3] <> '\') then begin
      MsgBoxP (SetupMessages[msgInvalidPath], nil, mbError, MB_OK);
      Exit;
    end;

    { Check for invalid characters after x:\ }
    {$IFDEF WIN32}
    if UsingWindows4 then
      BadDirChars := BadDirChars4
    else
      BadDirChars := BadDirChars3;
    {$ENDIF}
    for I := 4 to Length(T) do
      if Pos(T[I], BadDirChars) <> 0 then begin
        MsgBox (FmtSetupMessage1({$IFDEF WIN32} msgBadDirName32 {$ELSE} msgBadDirName16 {$ENDIF},
          SpaceString(BadDirChars)), '', mbError, MB_OK);
        Exit;
      end;

    { Check if it's a valid drive }
    if not(UpCase(T[1]) in ['A'..'Z']) or not GetTotalFreeSpace(T[1], TotalFreeSpace) then begin
      MsgBoxP (SetupMessages[msgInvalidDrive], nil, mbError, MB_OK);
      Exit;
    end;

    { Remove trailing backslash }
    T := RemoveBackslashUnlessRoot(T);
    DirEdit.Text := T;

    {$IFNDEF WIN32}
    { On 16-bit installations, check to make the sure the path name is in
      8.3 format }
    if BadPath(Copy(T, 4, Maxint)) then begin
      MsgBoxP (SetupMessages[msgPathTooLong], nil, mbError, MB_OK);
      Exit;
    end;
    {$ENDIF}

    { Check if enough disk space (using an unsigned long compare) }
    {$IFDEF WIN32}
    if Cardinal(TotalFreeSpace) < Cardinal(MinimumSpace) then
    {$ELSE}
    if (LongRec(TotalFreeSpace).Hi < LongRec(MinimumSpace).Hi) or
       ((LongRec(TotalFreeSpace).Hi = LongRec(MinimumSpace).Hi) and
        (LongRec(TotalFreeSpace).Lo < LongRec(MinimumSpace).Lo)) then
    {$ENDIF}
      { If not, show warning }
      if MsgBox(FmtSetupMessage(msgDiskSpaceWarning,
           [IntToStr(MinimumSpace div 1024), IntToStr(TotalFreeSpace div 1024)]),
         StrPas(SetupMessages[msgDiskSpaceWarningTitle]),
         mbConfirmation, MB_YESNO or MB_DEFBUTTON2) <> ID_YES then
        Exit;

    { Check if directory already exists }
    if not(shDisableDirExistsWarning in SetupHeader.Options) and
       DirExists(T) then
      { If so, ask if user wants to install there anyway }
      if MsgBox(FmtSetupMessage1(msgDirExists, T),
        StrPas(SetupMessages[msgDirExistsTitle]),
        mbConfirmation, MB_YESNO) <> ID_YES then Exit;

    { Check if directory *doesn't* already exist }
    if (shEnableDirDoesntExistWarning in SetupHeader.Options) and
       not DirExists(T) then
      { If not, ask if user wants to install there anyway }
      if MsgBox(FmtSetupMessage1(msgDirDoesntExist, T),
        StrPas(SetupMessages[msgDirDoesntExistTitle]),
        mbConfirmation, MB_YESNO) <> ID_YES then Exit;

    Result := True;
  end;
  function CheckSelectProgramGroupPage: Boolean;
  const
    BadGroupCharsOld = '\/:*?"<>|,';
    BadGroupCharsNew = '/:*?"<>|,';
  var
    T, BadGroupChars: String;
    I: Integer;
  begin
    Result := False;

    if not NoIconsCheck.Checked then begin
      { Remove any leading or trailing spaces and out of place backslashes }
      T := Trim(GroupEdit.Text);
      {$IFDEF WIN32}
      I := Pos('\\', T);
      while I <> 0 do begin
        Delete (T, I, 1);
        I := Pos('\\', T);
      end;
      if (T <> '') and (T[1] = '\') then
        Delete (T, 1, 1);
      T := RemoveBackslash(T);
      {$ENDIF}
      GroupEdit.Text := T;

      if T = '' then begin
        MsgBoxP (SetupMessages[msgMustEnterGroupName], nil, mbError, MB_OK);
        Exit;
      end;

      {$IFDEF WIN32}
      if not UsingWindows4 then
      {$ENDIF}
        BadGroupChars := BadGroupCharsOld
      {$IFDEF WIN32}
      else
        BadGroupChars := BadGroupCharsNew
      {$ENDIF};

      { Check for invalid characters. Quotes are not allowed because the
        PROGMAN DDE interface uses them to delimit command strings. And
        since Windows 95/NT 4.0 stores program groups as literal directory
        names, quotes and the rest of the characters in the BadGroupChars
        constant aren't allowed either. }
      for I := 1 to Length(T) do
        if Pos(T[I], BadGroupChars) <> 0 then begin
          MsgBox (FmtSetupMessage1(msgBadGroupName, SpaceString(BadGroupChars)),
            '', mbError, MB_OK);
          Exit;
        end;
    end;

    Result := True;
  end;
begin
  case CurPage of
    wpPassword: if not CheckPassword then Exit;
    wpSelectDir: if not CheckSelectDirPage then Exit;
    wpSelectProgramGroup: if not CheckSelectProgramGroupPage then Exit;
  end;

  { Go to the next page, or close wizard if it was on the last page }
  if CurPage in [wpReady, wpFinished] then begin
    CantShow := True;
    DisableCloseQuery := True;
    Close;
    DisableCloseQuery := False;
    PostMessage (MainForm.Handle, WM_SETUPNEXTSTEP, 0, 0);
  end
  else begin
    Inc (CurPage);
    while SkipCurPage do
      Inc (CurPage);
    CurPageChanged;
  end;
end;

procedure TWizardForm.BackButtonClick(Sender: TObject);
begin
  if CurPage = Low(TWizardPage) then Exit;

  { Go to the previous page }
  Dec (CurPage);
  while SkipCurPage do
    Dec (CurPage);
  CurPageChanged;
end;

procedure TWizardForm.CancelButtonClick(Sender: TObject);
begin
  { Clicking Cancel will do the same thing as the Close button }
  Close;
end;

procedure TWizardForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  if not DisableCloseQuery then begin
    DisableCloseQuery := True;
    try
      { Redirect an attempt to close this form to MainForm }
      MainForm.Close;
      { And only close if MainForm closed }
      CanClose := MainForm.Closing;
    finally
      DisableCloseQuery := False;
    end;
  end;
end;

procedure TWizardForm.ReadGroupList;
{ Fills in the list box of existing groups }
var
  PROGMAN, Item: TDDEString;
  Conv: TDDEConversation;
  DataHandle: TDDEData;
  DataPointer: PChar;
  DataSize: LongInt;
  {$IFDEF WIN32}
  H: THandle;
  FindData: TWin32FindData;
  S: String;
  {$ENDIF}
begin
  try
    {$IFDEF WIN32}
    if not UsingWindows4 then begin
    {$ENDIF}
      PROGMAN := DDE.CreateString('PROGMAN');
      Item := DDE.CreateString('GROUPS');
      Conv := DDE.BeginConnection(PROGMAN, PROGMAN);
      try
        DDE.RequestBegin (Conv, Item, CF_TEXT, DataHandle, Pointer(DataPointer), DataSize);
        try
          GroupList.Items.SetText (DataPointer);
        finally
          DDE.RequestEnd (DataHandle);
        end;
      finally
        DDE.EndConnection (Conv);
        DDE.FreeString (Item);
        DDE.FreeString (PROGMAN);
      end;
    {$IFDEF WIN32}
    end
    else begin
      H := FindFirstFile(PChar(ChangeDirConst('{group}\*')), FindData);
      if H <> INVALID_HANDLE_VALUE then begin
        repeat
          if FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY <> 0 then begin
            S := FindData.cFileName;
            if (S <> '.') and (S <> '..') then
              GroupList.Items.Add (S);
          end;
        until not FindNextFile(H, FindData);
        WinProcs.FindClose (H);
      end;
    end;
    {$ENDIF}
  except
    { ignore any exceptions }
  end;
end;

procedure TWizardForm.DirListChange(Sender: TObject);
begin
  if not(shDisableAppendDir in SetupHeader.Options) then
    DirEdit.Text := AddBackslash(DirList.Directory) +
      ExtractFileName(StrPas(SetupHeader.DefaultDirName))
  else
    DirEdit.Text := DirList.Directory;
end;

procedure TWizardForm.GroupEditChange(Sender: TObject);
var
  T: String;
  I: Integer;
  SaveOnClick: TNotifyEvent;
label 1;
begin
  T := GroupEdit.Text;
  with GroupList do begin
    for I := 0 to Items.Count-1 do
      if AnsiCompareText(T, Items[I]) = 0 then
        goto 1;
    I := -1;
  1:if ItemIndex <> I then begin
      SaveOnClick := OnClick;
      { Temporarily disabled the OnChange event }
      OnClick := nil;
      try
        ItemIndex := I;
      finally
        OnClick := SaveOnClick;
      end;
    end;
  end;
end;

procedure TWizardForm.GroupListClick(Sender: TObject);
var
  SaveOnChange: TNotifyEvent;
begin
  with GroupList do
    if ItemIndex <> -1 then begin
      SaveOnChange := GroupEdit.OnChange;
      { Temporarily disabled the OnChange event }
      GroupEdit.OnChange := nil;
      try
        GroupEdit.Text := Items[ItemIndex];
      finally
        GroupEdit.OnChange := SaveOnChange;
      end;
    end;
end;

procedure TWizardForm.NoIconsCheckClick(Sender: TObject);
const
  ColorChange: array[Boolean] of TColor = (clBtnFace, clWindow);
begin
  with GroupEdit do begin
    Enabled := not NoIconsCheck.Checked;
    Color := ColorChange[Enabled];
  end;
  with GroupList do begin
    Enabled := not NoIconsCheck.Checked;
    Color := ColorChange[Enabled];
    if not Enabled then ItemIndex := -1;
  end;
end;

procedure TWizardForm.FormShow(Sender: TObject);
  function TrySetDirListDirectory (const S: String): Boolean;
  begin
    Result := True;
    try
      DirList.Directory := S;
    except
      Result := False;
    end;
  end;
var
  P, P2, P3: String;
begin
  if ShowInitDone then Exit;
  ShowInitDone := True;

  if shCreateAppDir in SetupHeader.Options then begin
    { Try different values and see what works. The 'except end;' throws away any
      exceptions that might occur. }
    TrySetDirListDirectory ('\');
    TrySetDirListDirectory (SystemDrive + '\');
    P := ChangeDirConstPChar(SetupHeader.DefaultDirName);
    if not(shDisableAppendDir in SetupHeader.Options) or
       not TrySetDirListDirectory(P) then begin
      P2 := P;
      while True do begin
        P3 := RemoveBackslashUnlessRoot(ExtractFilePath(P2));
        if Length(P3) < 3 then Break;
        if TrySetDirListDirectory(P3) or (P3 = P2) then
          Break;
        P2 := P3;
      end;
    end;
    DirEdit.Text := P;

    if (InitDir <> '') and not(shDisableDirPage in SetupHeader.Options) then begin
      { ^ InitDir currently isn't supported for installations with
        shDisableDirPage set. If the wizard page isn't displayed, it doesn't
        get a chance to validate the path specified. }
      try
        DirList.Directory := ExtractFilePath(InitDir);
        DirList.Directory := InitDir;
      except
      end;
      DirEdit.Text := InitDir;
    end;
  end
  else
    DirEdit.Text := WinDir;
  if (InitProgramGroup <> '') and not(shDisableProgramGroupPage in SetupHeader.Options) then
    { ^ InitProgramGroup currently isn't supported for installations with
      shDisableProgramGroupPage set. If the wizard page isn't displayed, it
      doesn't get a chance to validate the program group name specified. }
    GroupEdit.Text := InitProgramGroup
  else
    GroupEdit.Text := StrPas(SetupHeader.DefaultGroupName);

  NoIconsCheck.Visible := shAllowNoIcons in SetupHeader.Options;

  CurPage := Low(TWizardPage);
  CurPageChanged;

  ReadGroupList;
end;

procedure TWizardForm.PicturePaintBoxPaint(Sender: TObject);
var
  R: TRect;
  X, Y: Integer;
begin
  with PicturePaintBox do begin
    R := ClientRect;
    { Draw the "sunken" edge, and exclude the edge from the clipping region }
    NewDrawEdge (Canvas, R, deThinSunken);
    InflateRect (R, -1, -1);
    IntersectClipRect (Canvas.Handle, R.Left, R.Top, R.Right, R.Bottom);
    { Draw the teal background }
    Canvas.Brush.Color := SetupHeader.WizardImageBackColor;
    Canvas.FillRect (R);
    { And finally draw the image }
    if Assigned(WizardImage) then begin
      X := R.Left + ((R.Right - R.Left) - WizardImage.Width) div 2;
      if X < 0 then X := 0;
      Y := R.Top + ((R.Bottom - R.Top) - WizardImage.Height) div 2;
      if Y < 0 then Y := 0;
      Canvas.Draw (X, Y, WizardImage);
    end;
  end;
end;

procedure TWizardForm.WelcomeBoxPaint(Sender: TObject);
var
  {$IFNDEF WIN32}
  Buf: array[0..4095] of Char;
  {$ELSE}
  R: TRect;
  {$ENDIF}
begin
  { Uses a paint box instead of a label because in Delphi 1 labels are
    limited to 255 characters (even if you use SetTextBuf) }
  {$IFNDEF WIN32}
  FixLabelBuf (Buf, SizeOf(Buf)-1, SetupMessages[msgWelcomeLabel]);
  StrLCat (Buf, SNewLine2, SizeOf(Buf)-1);
  StrLCat (Buf, SetupMessages[msgClickNext], SizeOf(Buf)-1);
  DrawText (WelcomeBox.Canvas.Handle, Buf, -1, WelcomeBox.ClientRect,
    DT_LEFT or DT_NOPREFIX or DT_WORDBREAK);
  {$ELSE}
  { A more optimized version for Delphi 2 & up, since it doesn't have the
    255 character string length limit }
  R := WelcomeBox.ClientRect;
  DrawText (WelcomeBox.Canvas.Handle,
    PChar(FixLabel(SetupMessages[msgWelcomeLabel] +
      SNewLine2 + SetupMessages[msgClickNext])), -1, R,
    DT_LEFT or DT_NOPREFIX or DT_WORDBREAK);
  {$ENDIF}
end;

procedure TWizardForm.SetupIconPaint(Sender: TObject);
begin
  (Sender as TPaintBox).Canvas.Draw (0, 0, Application.Icon);
end;

procedure TWizardForm.WMSysCommand (var Message: TWMSysCommand);
begin
  if Message.CmdType and $FFF0 = SC_MINIMIZE then
    { A minimize button is shown on the wizard form when (shWindowVisible in
      SetupHeader.Options). When it is clicked we want to minimize the whole
      application. }
    Application.Minimize
  else
  if Message.CmdType = 9999 then
    MainForm.ShowAboutBox
  else
    inherited;
end;

end.
